"""Tests for RestraintsParameterisation and associated classes used in refinement"""

from __future__ import annotations

import os
import random

import pytest

from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx.phil import parse

from dials.algorithms.refinement import RefinerFactory
from dials.algorithms.refinement.restraints import RestraintsParameterisation
from dials.array_family import flex


def test_single_crystal_restraints_gradients():
    """Simple test with a single triclinic crystal restrained to a target unit cell"""

    from dxtbx.model.experiment_list import Experiment, ExperimentList

    from dials.algorithms.refinement.parameterisation.beam_parameters import (
        BeamParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.detector_parameters import (
        DetectorParameterisationSinglePanel,
    )
    from dials.algorithms.refinement.parameterisation.prediction_parameters import (
        XYPhiPredictionParameterisation,
    )

    from . import geometry_phil
    from .setup_geometry import Extract

    overrides = """geometry.parameters.crystal.a.length.range = 10 50
  geometry.parameters.crystal.b.length.range = 10 50
  geometry.parameters.crystal.c.length.range = 10 50"""

    master_phil = parse(geometry_phil)

    models = Extract(master_phil, overrides)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    mybeam = models.beam

    # Build a mock scan for a 72 degree sequence
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 720),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(720)),
        deg=True,
    )

    # Create parameterisations of these models
    det_param = DetectorParameterisationSinglePanel(mydetector)
    s0_param = BeamParameterisation(mybeam, mygonio)
    xlo_param = CrystalOrientationParameterisation(mycrystal)
    xluc_param = CrystalUnitCellParameterisation(mycrystal)

    # Create an ExperimentList
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )

    # Build a prediction parameterisation
    pred_param = XYPhiPredictionParameterisation(
        experiments,
        detector_parameterisations=[det_param],
        beam_parameterisations=[s0_param],
        xl_orientation_parameterisations=[xlo_param],
        xl_unit_cell_parameterisations=[xluc_param],
    )

    # Build a restraints parameterisation
    rp = RestraintsParameterisation(
        detector_parameterisations=[det_param],
        beam_parameterisations=[s0_param],
        xl_orientation_parameterisations=[xlo_param],
        xl_unit_cell_parameterisations=[xluc_param],
    )

    # make a unit cell target
    sigma = 1.0
    uc = mycrystal.get_unit_cell().parameters()
    target_uc = [random.gauss(e, sigma) for e in uc]

    rp.add_restraints_to_target_xl_unit_cell(
        experiment_id=0, values=target_uc, sigma=[sigma] * 6
    )

    # get analytical values and gradients
    vals, grads, weights = rp.get_residuals_gradients_and_weights()
    assert len(vals) == rp.num_residuals

    # get finite difference gradients
    p_vals = pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    fd_grad = []

    for i, delta in enumerate(deltas):
        val = p_vals[i]

        p_vals[i] -= delta / 2.0
        pred_param.set_param_vals(p_vals)

        rev_state, foo, bar = rp.get_residuals_gradients_and_weights()
        rev_state = flex.double(rev_state)

        p_vals[i] += delta
        pred_param.set_param_vals(p_vals)

        fwd_state, foo, bar = rp.get_residuals_gradients_and_weights()
        fwd_state = flex.double(fwd_state)

        p_vals[i] = val

        fd = (fwd_state - rev_state) / delta
        fd_grad.append(fd)

    # for comparison, fd_grad is a list of flex.doubles, each of which corresponds
    # to a column of the sparse matrix grads.
    for i, fd in enumerate(fd_grad):
        # extract dense column from the sparse matrix
        an = grads.col(i).as_dense_vector()

        assert an == pytest.approx(fd, abs=1e-5)


def test_two_triclinic_crystals():
    """Simple test with two triclinic crystals restrained to a target unit cell"""

    from dxtbx.model.experiment_list import Experiment, ExperimentList

    from dials.algorithms.refinement.parameterisation.beam_parameters import (
        BeamParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.detector_parameters import (
        DetectorParameterisationSinglePanel,
    )

    #### Import model parameterisations
    from dials.algorithms.refinement.parameterisation.prediction_parameters import (
        XYPhiPredictionParameterisation,
    )

    from . import geometry_phil
    from .setup_geometry import Extract

    overrides = """geometry.parameters.crystal.a.length.range = 10 50
  geometry.parameters.crystal.b.length.range = 10 50
  geometry.parameters.crystal.c.length.range = 10 50"""

    master_phil = parse(geometry_phil)

    models = Extract(master_phil, overrides)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    # duplicate the crystal
    from copy import deepcopy

    mycrystal2 = deepcopy(mycrystal)
    mybeam = models.beam

    # Build a mock scan for a 72 degree sequence
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 720),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(720)),
        deg=True,
    )

    # Create parameterisations of these models
    det_param = DetectorParameterisationSinglePanel(mydetector)
    s0_param = BeamParameterisation(mybeam, mygonio)
    xlo_param = CrystalOrientationParameterisation(mycrystal)
    xluc_param = CrystalUnitCellParameterisation(mycrystal)
    xluc_param2 = CrystalUnitCellParameterisation(mycrystal2, experiment_ids=[1])

    # Create an ExperimentList with the crystal duplicated
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal2,
            imageset=None,
        )
    )

    # Build a prediction parameterisation
    pred_param = XYPhiPredictionParameterisation(
        experiments,
        detector_parameterisations=[det_param],
        beam_parameterisations=[s0_param],
        xl_orientation_parameterisations=[xlo_param],
        xl_unit_cell_parameterisations=[xluc_param, xluc_param2],
    )

    # Build a restraints parameterisation
    rp = RestraintsParameterisation(
        detector_parameterisations=[det_param],
        beam_parameterisations=[s0_param],
        xl_orientation_parameterisations=[xlo_param],
        xl_unit_cell_parameterisations=[xluc_param, xluc_param2],
    )

    # make a unit cell target
    sigma = 1.0
    uc = mycrystal.get_unit_cell().parameters()
    target_uc = [random.gauss(e, sigma) for e in uc]

    rp.add_restraints_to_target_xl_unit_cell(
        experiment_id=0, values=target_uc, sigma=[sigma] * 6
    )
    rp.add_restraints_to_target_xl_unit_cell(
        experiment_id=1, values=target_uc, sigma=[sigma] * 6
    )

    # get analytical values and gradients
    vals, grads, weights = rp.get_residuals_gradients_and_weights()
    assert len(vals) == rp.num_residuals

    # get finite difference gradients
    p_vals = pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    fd_grad = []

    for i, delta in enumerate(deltas):
        val = p_vals[i]

        p_vals[i] -= delta / 2.0
        pred_param.set_param_vals(p_vals)

        rev_state, foo, bar = rp.get_residuals_gradients_and_weights()
        rev_state = flex.double(rev_state)

        p_vals[i] += delta
        pred_param.set_param_vals(p_vals)

        fwd_state, foo, bar = rp.get_residuals_gradients_and_weights()
        fwd_state = flex.double(fwd_state)

        p_vals[i] = val

        fd = (fwd_state - rev_state) / delta
        fd_grad.append(fd)

    # for comparison, fd_grad is a list of flex.doubles, each of which corresponds
    # to a column of the sparse matrix grads.
    for i, fd in enumerate(fd_grad):
        # extract dense column from the sparse matrix
        an = grads.col(i).as_dense_vector()
        assert an == pytest.approx(fd, abs=1e-5)


def test_10_crystals_with_stills_parameterisation(dials_regression):
    """Test with multiple crystals, and a stills refiner"""

    # The phil scope
    from dials.algorithms.refinement.refiner import phil_scope

    user_phil = parse(
        """
  refinement
  {
    parameterisation
    {
      crystal
      {
        unit_cell
        {
          restraints
          {
            tie_to_target
            {
              values=95,95,132,90,90,120
              sigmas=1,1,1,0,0,0
              id=0,1,2,3,4,5,6,7,8,9
            }
          }
        }
      }
    }
  }
  """
    )

    working_phil = phil_scope.fetch(source=user_phil)
    working_params = working_phil.extract()

    # use the multi stills test data
    data_dir = os.path.join(dials_regression, "refinement_test_data", "multi_stills")
    experiments_path = os.path.join(data_dir, "combined_experiments.json")
    pickle_path = os.path.join(data_dir, "combined_reflections.pickle")

    experiments = ExperimentListFactory.from_json_file(
        experiments_path, check_format=False
    )
    reflections = flex.reflection_table.from_file(pickle_path)

    refiner = RefinerFactory.from_parameters_data_experiments(
        working_params, reflections, experiments
    )

    # hack to extract the objects needed from the Refiner
    rp = refiner._target._restraints_parameterisation
    pred_param = refiner._pred_param

    # get analytical values and gradients
    vals, grads, weights = rp.get_residuals_gradients_and_weights()
    assert len(vals) == rp.num_residuals

    # get finite difference gradients
    p_vals = pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    fd_grad = []

    for i, delta in enumerate(deltas):
        val = p_vals[i]

        p_vals[i] -= delta / 2.0
        pred_param.set_param_vals(p_vals)

        rev_state, foo, bar = rp.get_residuals_gradients_and_weights()
        rev_state = flex.double(rev_state)

        p_vals[i] += delta
        pred_param.set_param_vals(p_vals)

        fwd_state, foo, bar = rp.get_residuals_gradients_and_weights()
        fwd_state = flex.double(fwd_state)

        p_vals[i] = val

        fd = (fwd_state - rev_state) / delta
        fd_grad.append(fd)

    # for comparison, fd_grad is a list of flex.doubles, each of which corresponds
    # to a column of the sparse matrix grads.
    pnames = pred_param.get_param_names()
    for i, (pname, fd) in enumerate(zip(pnames, fd_grad)):
        # extract dense column from the sparse matrix
        an = grads.col(i).as_dense_vector()

        # print pname
        # print list(an.round(6))
        # print list(fd.round(6))
        # print
        assert an == pytest.approx(fd, abs=1e-5)


def test_group_restraint_with_multiple_crystals_and_a_stills_refiner(dials_regression):

    # The phil scope
    from dials.algorithms.refinement.refiner import phil_scope

    user_phil = parse(
        """
  refinement
  {
    parameterisation
    {
      crystal
      {
        unit_cell
        {
          restraints
          {
            tie_to_group
            {
              sigmas=1,0,2,0,0,0
            }
          }
        }
      }
    }
  }
  """
    )

    working_phil = phil_scope.fetch(source=user_phil)
    working_params = working_phil.extract()

    # use the multi stills test data
    data_dir = os.path.join(dials_regression, "refinement_test_data", "multi_stills")
    experiments_path = os.path.join(data_dir, "combined_experiments.json")
    pickle_path = os.path.join(data_dir, "combined_reflections.pickle")

    experiments = ExperimentListFactory.from_json_file(
        experiments_path, check_format=False
    )
    reflections = flex.reflection_table.from_file(pickle_path)

    refiner = RefinerFactory.from_parameters_data_experiments(
        working_params, reflections, experiments
    )

    # hack to extract the objects needed from the Refiner
    rp = refiner._target._restraints_parameterisation
    pred_param = refiner._pred_param

    # get analytical values and gradients
    vals, grads, weights = rp.get_residuals_gradients_and_weights()
    assert len(vals) == rp.num_residuals

    # get finite difference gradients
    p_vals = pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    fd_grad = []

    for i, delta in enumerate(deltas):
        val = p_vals[i]

        p_vals[i] -= delta / 2.0
        pred_param.set_param_vals(p_vals)

        rev_state, foo, bar = rp.get_residuals_gradients_and_weights()
        rev_state = flex.double(rev_state)

        p_vals[i] += delta
        pred_param.set_param_vals(p_vals)

        fwd_state, foo, bar = rp.get_residuals_gradients_and_weights()
        fwd_state = flex.double(fwd_state)

        p_vals[i] = val

        fd = (fwd_state - rev_state) / delta
        fd_grad.append(fd)

    # for comparison, fd_grad is a list of flex.doubles, each of which corresponds
    # to the gradients of the residuals wrt to a single parameter.
    pnames = pred_param.get_param_names()
    for i, (pname, fd) in enumerate(zip(pnames, fd_grad)):
        # extract dense column from the sparse matrix
        an = grads.col(i).as_dense_vector()

        # print pname
        # print list(an.round(6))
        # print list(fd.round(6))
        # print
        assert an == pytest.approx(fd, abs=1e-5)



from __future__ import annotations

import pytest

from cctbx import crystal

from dials.algorithms.indexing.ssx.analysis import (
    combine_results_dicts,
    generate_html_report,
    generate_plots,
    make_cluster_plots,
    make_summary_table,
)


def generate_test_results_dict(n_lattices=1):
    results = {
        0: [
            {
                "Image": "test_image_001.cbf",
                "n_indexed": 0,
                "n_strong": 100,
            }
        ],  # an unindexed image
        1: [
            {
                "Image": "test_image_002.cbf",
                "n_indexed": 50,
                "n_strong": 200,
                "RMSD_X": 0.5,
                "RMSD_Y": 1.1,
                "RMSD_dPsi": 1.2,
            }
        ],  # an indexed image with one lattice
        2: [
            {
                "Image": "test_image_003.cbf",
                "n_indexed": 30,
                "n_strong": 50,
                "RMSD_X": 0.5,
                "RMSD_Y": 0.4,
                "RMSD_dPsi": 0.6,
            }
        ],  # an indexed image with one lattice
    }
    if n_lattices > 1:
        results[2].append(
            {
                "Image": "test_image_003.cbf",
                "n_indexed": 10,
                "n_strong": 50,
                "RMSD_X": 0.3,
                "RMSD_Y": 0.5,
                "RMSD_dPsi": 0.7,
            }
        )  # an indexed image with two lattices
    return results


@pytest.mark.parametrize("n_lattices", [1, 2])
def test_make_summary_table(n_lattices):
    """Test that the summary table has the correct columns"""
    results = generate_test_results_dict(n_lattices)
    table = make_summary_table(results)
    headerline = table.splitlines()[1]
    headers = ["Image", "expt_id", "n_indexed", "RMSD X", "RMSD Y", "RMSD dPsi"]
    last_lattice_line = table.splitlines()[-2]
    assert all(h in headerline for h in headers)
    if n_lattices > 1:
        assert "lattice" in headerline
        expected_last = "| test_image_003.cbf |         2 |         2 | 10/50 (20.0%)  |      0.3 |      0.5 |         0.7 |"
    else:
        assert "lattice" not in headerline
        expected_last = "| test_image_003.cbf |         1 | 30/50 (60.0%)  |      0.5 |      0.4 |         0.6 |"
    assert expected_last == last_lattice_line


def test_combine_multiple_results_summary_dicts():
    s1 = generate_test_results_dict()
    s2 = generate_test_results_dict(n_lattices=2)
    combined = combine_results_dicts([s1, s2])
    assert list(combined.keys()) == [0, 1, 2, 3, 4, 5]
    for k, res in combined.items():
        if k == 5:
            assert len(res) == 2
        else:
            assert len(res) == 1


@pytest.mark.parametrize("n_lattices", [1, 2])
def test_generate_plots(n_lattices):
    plots = generate_plots(generate_test_results_dict(n_lattices))

    # First plot, n_indexed vs image, should have a scatter plot n_strong,
    # one for the first lattice an another for second lattices
    assert len(plots["n_indexed"]["data"]) == n_lattices + 1
    assert len(plots["n_indexed"]["data"][0]["x"]) == 3  # lattice 1 (0 or indexed)
    assert plots["n_indexed"]["data"][0]["y"][0] == 0.0  # first has none indexed
    assert all(plots["n_indexed"]["data"][0]["y"][i] > 0.0 for i in range(1, 3))
    assert plots["n_indexed"]["data"][-1]["y"] == [100, 200, 50]  # n_strong

    if n_lattices == 2:
        assert (
            len(plots["n_indexed"]["data"][1]["x"]) == 1
        )  # only one has second lattice
        assert "lattice 1" in plots["n_indexed"]["data"][0]["name"]
        assert "lattice 2" in plots["n_indexed"]["data"][1]["name"]
        percent_indexed = [0.0, 25.0, 80.0]
    else:
        assert "lattice 1" not in plots["n_indexed"]["data"][0]["name"]
        percent_indexed = [0.0, 25.0, 60.0]

    # Next plot, percentage of strong spots, should just have one trend
    assert plots["percent_indexed"]["data"][0]["y"] == percent_indexed

    # Perecent indexed histogram
    assert sum(plots["percent_indexed_hist"]["data"][0]["y"]) == 3

    # rmsd plots
    assert plots["rmsds"]["data"][0]["y"] == [0.5, 0.5]  # X
    assert plots["rmsds"]["data"][1]["y"] == [1.1, 0.4]  # Y
    assert plots["rmsdz"]["data"][0]["y"] == [1.2, 0.6]  # dPsi
    if n_lattices == 2:
        assert plots["rmsds"]["data"][2]["y"] == [0.3]
        assert plots["rmsds"]["data"][3]["y"] == [0.5]
        assert plots["rmsdz"]["data"][1]["y"] == [0.7]  # dPsi

    # rmsd distribution plots
    assert sum(plots["rmsdxy_hist"]["data"][0]["y"]) == n_lattices + 1
    assert len(plots["rmsdxy_hist"]["data"]) == 2
    assert sum(plots["rmsdz_hist"]["data"][0]["y"]) == n_lattices + 1


def test_generate_html_report(tmp_path):
    plots = generate_plots(generate_test_results_dict())
    fname = "test_report_name.html"
    generate_html_report(plots, tmp_path / fname)
    assert tmp_path.joinpath("test_report_name.html").is_file()


def test_make_cluster_plots():
    from dials.algorithms.clustering.unit_cell import Cluster

    c1 = Cluster(
        [
            crystal.symmetry(
                unit_cell=(10.0, 10.0, 10.0, 90, 90, 90), space_group="P1"
            ),
            crystal.symmetry(
                unit_cell=(10.1, 10.1, 10.1, 90, 90, 90), space_group="P1"
            ),
            crystal.symmetry(
                unit_cell=(10.2, 10.2, 10.2, 90, 90, 90), space_group="P1"
            ),
        ]
    )
    c2 = Cluster(
        [
            crystal.symmetry(
                unit_cell=(11.0, 11.0, 11.0, 90, 90, 90), space_group="P1"
            ),
            crystal.symmetry(
                unit_cell=(11.1, 11.1, 11.1, 90, 90, 90), space_group="P1"
            ),
            crystal.symmetry(
                unit_cell=(11.2, 11.2, 11.2, 90, 90, 90), space_group="P1"
            ),
            crystal.symmetry(
                unit_cell=(11.3, 11.3, 11.3, 90, 90, 90), space_group="P1"
            ),
        ]
    )
    clusters = [c1, c2]
    plots = make_cluster_plots(clusters)
    assert "uc_scatter_0" in plots
    assert "uc_scatter_1" in plots
    assert "uc_hist_0" in plots
    assert "uc_hist_1" in plots
    print(plots)
    assert len(plots["uc_hist_0"]["data"]) == 3
    assert len(plots["uc_hist_0"]["data"][0]["x"]) == 3
    assert len(plots["uc_hist_1"]["data"][0]["x"]) == 4
    assert len(plots["uc_scatter_0"]["data"]) == 3
    assert len(plots["uc_scatter_0"]["data"][0]["x"]) == 3
    assert len(plots["uc_scatter_1"]["data"][0]["x"]) == 4



from __future__ import annotations

import logging

import scitbx.matrix
from rstbx.dps_core.cell_assessment import unit_cell_too_small
from rstbx.indexing_api import tools
from scitbx.array_family import flex

logger = logging.getLogger(__name__)


def correct(experiments, reflections, assign_indices, threshold=0.9):
    assert len(experiments.crystals()) == 1
    while True:
        sel = reflections["miller_index"] != (0, 0, 0)
        if sel.count(True) == 0:
            break
        T = detect(reflections["miller_index"].select(sel), threshold=threshold)
        if T is None:
            break

        crystal_model = experiments.crystals()[0]
        direct_matrix = scitbx.matrix.sqr(crystal_model.get_A()).inverse()
        M = T.inverse().transpose()
        new_direct_matrix = M * direct_matrix

        crystal_model.set_A(new_direct_matrix.inverse())

        unit_cell_too_small(crystal_model.get_unit_cell())
        cb_op = crystal_model.get_unit_cell().change_of_basis_op_to_niggli_cell()
        crystal_model.update(crystal_model.change_basis(cb_op))

        reflections["id"] = flex.int(len(reflections), -1)
        reflections.unset_flags(
            flex.bool(len(reflections), True), reflections.flags.indexed
        )
        assign_indices(reflections, experiments)


def detect(miller_indices, threshold=0.9):

    for test in tools.R:
        cum = tools.cpp_absence_test(miller_indices, test["mod"], test["vec"])
        for counter in range(test["mod"]):
            if float(cum[counter]) / miller_indices.size() > threshold and counter == 0:
                # (if counter != 0 there is no obvious way to correct this)
                logger.debug(
                    "Detected exclusive presence of %dH %dK %dL = %dn, remainder %d",
                    test["vec"][0],
                    test["vec"][1],
                    test["vec"][2],
                    test["mod"],
                    counter,
                )
                logger.debug(
                    "%s, %s, %s",
                    test["vec"],
                    test["mod"],
                    float(cum[counter]) / miller_indices.size(),
                )
                return test["trans"]



from __future__ import annotations

try:
    import boost_adaptbx.boost.python
except Exception:
    ext = None
else:
    ext = boost_adaptbx.boost.python.import_ext("recviewer_ext", optional=False)

if ext is not None:
    from recviewer_ext import *



from __future__ import annotations

import random

import pytest


def test():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import index_of_dispersion_filter

    # Create an image
    image = flex.random_double(2000 * 2000)
    image.reshape(flex.grid(2000, 2000))
    mask = flex.random_bool(2000 * 2000, 0.99).as_int()
    mask.reshape(flex.grid(2000, 2000))

    # Calculate the summed area table
    mask2 = mask.deep_copy()
    index_of_dispersion_filter = index_of_dispersion_filter(image, mask2, (3, 3), 2)
    mean = index_of_dispersion_filter.mean()
    var = index_of_dispersion_filter.sample_variance()
    index_of_dispersion = index_of_dispersion_filter.index_of_dispersion()

    # For a selection of random points, ensure that the value is the
    # sum of the area under the kernel
    eps = 1e-7
    for i in range(10000):
        i = random.randint(10, 1990)
        j = random.randint(10, 1990)
        m1 = mean[j, i]
        v1 = var[j, i]
        f1 = index_of_dispersion[j, i]
        p = image[j - 3 : j + 4, i - 3 : i + 4]
        m = mask[j - 3 : j + 4, i - 3 : i + 4]
        if mask[j, i] == 0:
            m2 = 0.0
            v2 = 0.0
            f2 = 1.0
        else:
            p = flex.select(p, flags=m)
            mv = flex.mean_and_variance(flex.double(p))
            m2 = mv.mean()
            v2 = mv.unweighted_sample_variance()
            f2 = v2 / m2
        assert m1 == pytest.approx(m2, abs=eps)
        assert v1 == pytest.approx(v2, abs=eps)
        assert f1 == pytest.approx(f2, abs=eps)



from __future__ import annotations

import os
import pathlib
import shutil

import procrunner
import pytest

from dxtbx.serialize import load


@pytest.mark.parametrize("use_beam", ["True", "False"])
@pytest.mark.parametrize("use_gonio", ["True", "False"])
@pytest.mark.parametrize("use_detector", ["True", "False"])
def test_reference_individual(dials_data, tmpdir, use_beam, use_gonio, use_detector):

    expected_beam = {"True": 3, "False": 0.9795}
    expected_gonio = {
        "True": (7, 8, 9, 4, 5, 6, 1, 2, 3),
        "False": (1, 0, 0, 0, 1, 0, 0, 0, 1),
    }
    expected_detector = {"True": "Fake panel", "False": "Panel"}

    # Find the image files
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )

    # Create an experiment with some faked geometry items
    fake_phil = """
      geometry {
        beam.wavelength = 3
        detector.panel.name = "Fake panel"
        goniometer.fixed_rotation = 7,8,9,4,5,6,1,2,3
      }
  """
    tmpdir.join("fake.phil").write(fake_phil)
    fake_result = procrunner.run(
        ["dials.import", "fake.phil", "output.experiments=fake_geometry.expt"]
        + image_files,
        working_directory=tmpdir,
    )
    assert not fake_result.returncode and not fake_result.stderr
    assert tmpdir.join("fake_geometry.expt").check(file=1)

    # Write an import phil file
    import_phil = f"""
      input {{
        reference_geometry = fake_geometry.expt
        check_reference_geometry = False
        use_beam_reference = {use_beam}
        use_gonio_reference = {use_gonio}
        use_detector_reference = {use_detector}
      }}
  """
    tmpdir.join("test_reference_individual.phil").write(import_phil)

    result = procrunner.run(
        [
            "dials.import",
            "test_reference_individual.phil",
            "output.experiments=reference_geometry.expt",
        ]
        + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("reference_geometry.expt").check(file=1)

    experiments = load.experiment_list(tmpdir.join("reference_geometry.expt"))
    assert experiments[0].identifier != ""
    imgset = experiments[0].imageset

    beam = imgset.get_beam()
    goniometer = imgset.get_goniometer()
    detector = imgset.get_detector()

    assert beam.get_wavelength() == expected_beam[use_beam]
    assert goniometer.get_fixed_rotation() == expected_gonio[use_gonio]
    assert detector[0].get_name() == expected_detector[use_detector]


def test_multiple_sequence_import_fails_when_not_allowed(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )
    del image_files[4]  # Delete filename to force two sequences

    # run without allowing multiple sequences
    result = procrunner.run(
        [
            "dials.import",
            "output.experiments=experiments_multiple_sequences.expt",
            "allow_multiple_sequence=False",
        ]
        + image_files,
        working_directory=tmpdir,
    )
    assert result.returncode == 1
    assert b"ore than 1 sequence" in result.stderr
    assert not tmpdir.join("experiments_multiple_sequences.expt").check()


def test_can_import_multiple_sequences(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )
    del image_files[4]  # Delete filename to force two sequences

    result = procrunner.run(
        ["dials.import", "output.experiments=experiments_multiple_sequences.expt"]
        + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("experiments_multiple_sequences.expt").check(file=1)

    experiments = load.experiment_list(
        tmpdir.join("experiments_multiple_sequences.expt")
    )
    assert len(experiments) == 2
    for experiment in experiments:
        assert experiment.identifier != ""


def test_with_mask(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )
    mask_filename = dials_data("centroid_test_data", pathlib=True) / "mask.pickle"

    result = procrunner.run(
        [
            "dials.import",
            f"mask={mask_filename}",
            "output.experiments=experiments_with_mask.expt",
        ]
        + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("experiments_with_mask.expt").check(file=1)

    experiments = load.experiment_list(tmpdir.join("experiments_with_mask.expt"))
    assert experiments[0].identifier != ""
    assert (
        pathlib.Path(experiments[0].imageset.external_lookup.mask.filename)
        == mask_filename
    )


def test_override_geometry(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )

    # Write a geometry phil file
    tmpdir.join("geometry.phil").write(
        """
      geometry {
        beam {
          wavelength = 2
          direction = (-1,0,0)
        }
        detector {
          panel {
            name = "New panel"
            type = "New type"
            pixel_size = 10,20
            image_size = 30,40
            trusted_range = 50,60
            thickness = 70
            material = "Si"
            fast_axis = -1,0,0
            slow_axis = 0,-1,0
            origin = 100,100,100
          }
        }
        goniometer {
          axes = 0,0,-1
          fixed_rotation = 0,1,2,3,4,5,6,7,8
          setting_rotation = 8,7,6,5,4,3,2,1,0
        }
        scan {
          image_range = 1,4
          oscillation = 1,2
        }
      }
  """
    )

    result = procrunner.run(
        ["dials.import", "geometry.phil", "output.experiments=override_geometry.expt"]
        + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("override_geometry.expt").check(file=1)

    experiments = load.experiment_list(tmpdir.join("override_geometry.expt"))
    assert experiments[0].identifier != ""
    imgset = experiments[0].imageset

    beam = imgset.get_beam()
    detector = imgset.get_detector()
    goniometer = imgset.get_goniometer()
    scan = imgset.get_scan()

    assert beam.get_wavelength() == 2
    assert beam.get_sample_to_source_direction() == (-1, 0, 0)
    assert detector[0].get_name() == "New panel"
    assert detector[0].get_type() == "New type"
    assert detector[0].get_pixel_size() == (10, 20)
    assert detector[0].get_image_size() == (30, 40)
    assert detector[0].get_trusted_range() == (50, 60)
    assert detector[0].get_thickness() == 70
    assert detector[0].get_material() == "Si"
    assert detector[0].get_fast_axis() == (-1, 0, 0)
    assert detector[0].get_slow_axis() == (0, -1, 0)
    assert detector[0].get_origin() == (100, 100, 100)
    assert goniometer.get_rotation_axis_datum() == (0, 0, -1)
    assert goniometer.get_fixed_rotation() == (0, 1, 2, 3, 4, 5, 6, 7, 8)
    assert goniometer.get_setting_rotation() == (8, 7, 6, 5, 4, 3, 2, 1, 0)
    assert scan.get_image_range() == (1, 4)
    assert scan.get_oscillation() == (1, 2)


def test_import_beam_centre(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )

    # provide mosflm beam centre to dials.import
    result = procrunner.run(
        [
            "dials.import",
            "mosflm_beam_centre=100,200",
            "output.experiments=mosflm_beam_centre.expt",
        ]
        + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("mosflm_beam_centre.expt").check(file=1)

    experiments = load.experiment_list(tmpdir.join("mosflm_beam_centre.expt"))
    imgset = experiments[0].imageset
    assert experiments[0].identifier != ""
    beam_centre = imgset.get_detector()[0].get_beam_centre(imgset.get_beam().get_s0())
    assert beam_centre == pytest.approx((200, 100))

    # provide an alternative models.expt to get geometry from
    result = procrunner.run(
        [
            "dials.import",
            "reference_geometry=mosflm_beam_centre.expt",
            "output.experiments=mosflm_beam_centre2.expt",
        ]
        + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("mosflm_beam_centre2.expt").check(file=1)
    experiments = load.experiment_list(tmpdir.join("mosflm_beam_centre2.expt"))
    imgset = experiments[0].imageset
    beam_centre = imgset.get_detector()[0].get_beam_centre(imgset.get_beam().get_s0())
    assert beam_centre == pytest.approx((200, 100))


def test_fast_slow_beam_centre(dials_regression, run_in_tmp_path):
    # test slow_fast_beam_centre with a multi-panel CS-PAD image
    impath = os.path.join(
        dials_regression,
        "image_examples",
        "LCLS_cspad_nexus",
        "idx-20130301060858401.cbf",
    )
    result = procrunner.run(
        [
            "dials.import",
            "fast_slow_beam_centre=42,134,18",
            "output.experiments=fast_slow_beam_centre.expt",
            impath,
        ]
    )
    assert not result.returncode and not result.stderr
    assert os.path.exists("fast_slow_beam_centre.expt")

    experiments = load.experiment_list("fast_slow_beam_centre.expt")
    imgset = experiments[0].imageset
    assert experiments[0].identifier != ""
    # beam centre on 18th panel
    s0 = imgset.get_beam().get_s0()
    beam_centre = imgset.get_detector()[18].get_beam_centre_px(s0)
    assert beam_centre == pytest.approx((42, 134))

    # check relative panel positions have not changed
    from scitbx import matrix

    o = matrix.col(imgset.get_detector()[0].get_origin())
    offsets = []
    for p in imgset.get_detector():
        intra_pnl = o - matrix.col(p.get_origin())
        offsets.append(intra_pnl.length())

    result = procrunner.run(
        ["dials.import", "output.experiments=reference.expt", impath]
    )
    assert not result.returncode and not result.stderr
    assert os.path.exists("reference.expt")

    ref_exp = load.experiment_list("reference.expt")
    ref_imset = ref_exp[0].imageset
    o = matrix.col(ref_imset.get_detector()[0].get_origin())
    ref_offsets = []
    for p in ref_imset.get_detector():
        intra_pnl = o - matrix.col(p.get_origin())
        ref_offsets.append(intra_pnl.length())
    assert offsets == pytest.approx(ref_offsets)


def test_from_image_files(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )

    # Import from the image files
    result = procrunner.run(
        ["dials.import", "output.experiments=imported.expt"] + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode
    assert tmpdir.join("imported.expt").check(file=1)
    # check that an experiment identifier is assigned
    exp = load.experiment_list(tmpdir.join("imported.expt"))
    assert exp[0].identifier != ""


def test_from_template(dials_data, tmpdir):
    # Find the image files
    templates = [
        dials_data("centroid_test_data", pathlib=True) / "centroid_####.cbf",
        dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf",
    ]

    for template in templates:
        # Import from the image files
        result = procrunner.run(
            [
                "dials.import",
                f"template={template}",
                "output.experiments=imported.expt",
            ],
            working_directory=tmpdir,
        )

        assert not result.returncode
        assert tmpdir.join("imported.expt").check(file=1)
        # check that an experiment identifier is assigned
        exp = load.experiment_list(tmpdir.join("imported.expt"))
        assert exp[0].identifier != ""


def test_extrapolate_scan(dials_data, tmpdir):
    # First image file
    image = dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf"

    result = procrunner.run(
        [
            "dials.import",
            image,
            "output.experiments=import_extrapolate.expt",
            "geometry.scan.image_range=1,900",
            "geometry.scan.extrapolate_scan=True",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode
    assert tmpdir.join("import_extrapolate.expt").check(file=1)
    # check that an experiment identifier is assigned
    exp = load.experiment_list(tmpdir.join("import_extrapolate.expt"))
    assert exp[0].identifier != ""


@pytest.fixture
def centroid_test_data_with_missing_image(dials_data, tmp_path):
    """
    Provide a testset with a missing image (#4) in a temporary directory
    (symlink if possible, copy if necessary), and clean up test files
    afterwards to conserve disk space.
    """
    images = sorted(
        dials_data("centroid_test_data", pathlib=True).glob(
            "centroid_*[1,2,3,5,6,7,8,9].cbf"
        )
    )
    for image in images:
        try:
            (tmp_path / image.name).symlink_to(image)
        except OSError:
            shutil.copy(image, tmp_path)
    yield tmp_path / "centroid_####.cbf"
    for image in images:
        try:
            (tmp_path / image.name).unlink()
        except PermissionError:
            pass


def test_template_with_missing_image_fails(centroid_test_data_with_missing_image):
    # This should fail because image #4 is missing
    for image_range in (None, (3, 5)):
        result = procrunner.run(
            ["dials.import", f"template={centroid_test_data_with_missing_image}"]
            + (["image_range=%i,%i" % image_range] if image_range else []),
            working_directory=centroid_test_data_with_missing_image.parent,
        )
        assert result.returncode
        assert b"Missing image 4 from imageset" in result.stderr


def test_template_with_missing_image_outside_of_image_range(
    centroid_test_data_with_missing_image,
):
    # Explicitly pass an image_range that doesn't include the missing image
    for image_range in ((1, 3), (5, 9)):
        result = procrunner.run(
            [
                "dials.import",
                f"template={centroid_test_data_with_missing_image}",
                "image_range=%i,%i" % image_range,
                "output.experiments=imported_%i_%i.expt" % image_range,
            ],
            working_directory=centroid_test_data_with_missing_image.parent,
        )
        assert not result.returncode
        expts = load.experiment_list(
            centroid_test_data_with_missing_image.parent.joinpath(
                "imported_%i_%i.expt" % image_range
            )
        )
        assert expts[0].scan.get_image_range() == image_range


def test_import_still_sequence_as_experiments(dials_data, tmp_path):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )

    out = "experiments_as_still.expt"

    procrunner.run(
        ["dials.import", "scan.oscillation=0,0", f"output.experiments={out}"]
        + image_files,
        working_directory=tmp_path,
    )

    imported_exp = load.experiment_list(tmp_path / out)
    assert len(imported_exp) == len(image_files)
    for exp in imported_exp:
        assert exp.identifier != ""

    iset = {exp.imageset for exp in imported_exp}
    assert len(iset) == 1

    # verify scans, goniometers kept too
    assert all(exp.scan.get_oscillation() == (0.0, 0.0) for exp in imported_exp)
    assert all(exp.goniometer is not None for exp in imported_exp)


def test_import_still_sequence_as_experiments_subset(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )[3:6]

    out = "experiments_as_still.expt"

    procrunner.run(
        ["dials.import", "scan.oscillation=10,0", f"output.experiments={out}"]
        + image_files,
        working_directory=tmpdir,
    )

    imported_exp = load.experiment_list(tmpdir.join(out))
    assert len(imported_exp) == len(image_files)
    for exp in imported_exp:
        assert exp.identifier != ""

    iset = {exp.imageset for exp in imported_exp}
    assert len(iset) == 1

    # verify scans, goniometers kept too
    assert all(exp.scan.get_oscillation() == (10.0, 0.0) for exp in imported_exp)
    assert all(exp.goniometer is not None for exp in imported_exp)


def test_import_still_sequence_as_expts_subset_by_range(dials_data, tmp_path):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )

    out = tmp_path / "experiments_as_still.expt"

    result = procrunner.run(
        [
            "dials.import",
            "scan.oscillation=10,0",
            "image_range=3,5",
            f"output.experiments={out}",
            *image_files,
        ],
        working_directory=tmp_path,
    )

    assert result.returncode == 0

    imported_exp = load.experiment_list(out)
    assert len(imported_exp) == 3
    for exp in imported_exp:
        assert exp.identifier != ""

    iset = {exp.imageset for exp in imported_exp}
    assert len(iset) == 1
    assert len(imported_exp[0].imageset) == 3

    assert list(iset)[0].get_image_identifier(0) == os.fspath(image_files[2])

    # verify scans, goniometers kept too
    assert all(exp.scan.get_oscillation() == (10.0, 0.0) for exp in imported_exp)
    assert all(exp.goniometer is not None for exp in imported_exp)


def test_import_still_sequence_as_experiments_split_subset(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )
    image_files = image_files[:3] + image_files[6:]

    out = "experiments_as_still.expt"

    procrunner.run(
        ["dials.import", "scan.oscillation=10,0", f"output.experiments={out}"]
        + image_files,
        working_directory=tmpdir,
    )

    imported_exp = load.experiment_list(tmpdir.join(out))
    assert len(imported_exp) == len(image_files)
    for exp in imported_exp:
        assert exp.identifier != ""

    iset = {exp.imageset for exp in imported_exp}
    assert len(iset) == 2


def test_with_convert_sequences_to_stills(dials_data, tmpdir):
    image_files = sorted(
        dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")
    )
    result = procrunner.run(
        [
            "dials.import",
            "convert_sequences_to_stills=True",
            "output.experiments=experiments_as_stills.expt",
        ]
        + image_files,
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("experiments_as_stills.expt").check(file=1)

    experiments = load.experiment_list(tmpdir.join("experiments_as_stills.expt"))
    for exp in experiments:
        assert exp.identifier != ""

    # should be no goniometers
    assert experiments.scans() == [None]
    assert experiments.goniometers() == [None]

    # should be same number of imagesets as images
    assert len(experiments.imagesets()) == len(image_files)

    # all should call out as still too
    assert experiments.all_stills()



from __future__ import annotations


class SimpleBackgroundExt:
    """An extension class implementing simple background subtraction."""

    name = "simple"

    @staticmethod
    def phil():
        from libtbx.phil import parse

        phil = parse(
            """
      outlier
        .help = "Outlier rejection prior to background fit"
      {
        algorithm = *null nsigma truncated normal plane tukey
          .help = "The outlier rejection algorithm."
          .type = choice

        nsigma
          .help = "Parameters for nsigma outlier rejector"
          .expert_level = 1
        {
          lower = 3
            .help = "Lower n sigma"
            .type = float
          upper = 3
            .help = "Upper n sigma"
            .type = float
        }

        truncated
          .help = "Parameters for truncated outlier rejector"
          .expert_level = 1
        {
          lower = 0.01
            .help = "Lower bound"
            .type = float
          upper = 0.01
            .help = "Upper bound"
            .type = float
        }

        normal
          .help = "Parameters for normal outlier rejector"
          .expert_level = 1
        {
          min_pixels = 10
            .help = "The minimum number of pixels to use in calculating the"
                    "background intensity."
            .type = int
        }

        plane
          .help = "Parameters for mosflm-like outlier rejector. This algorithm"
                  "is mainly used in conjunction with a linear 2d background."
          .expert_level = 1
        {
          fraction = 1.0
            .help = "The fraction of pixels to use in determining the initial"
                    "plane used for outlier rejection."
            .type = float

          n_sigma = 4.0
            .help = "The number of standard deviations above the threshold plane"
                    "to use in rejecting outliers from background calculation."
            .type = float
        }

        tukey
          .help = "Parameters for tukey outlier rejector"
          .expert_level = 1
        {
          lower = 1.5
            .help = "Lower IQR multiplier"
            .type = float
          upper = 1.5
            .help = "Upper IQR multiplier"
            .type = float
        }
      }

      model
        .help = "Background model"
      {
        algorithm = constant2d *constant3d linear2d linear3d
          .help = "The choice of background model"
          .type = choice
      }

      min_pixels = 10
        .type = int(value_min=1)
        .help = "The minimum number of pixels to compute the background"
    """
        )
        return phil

    def __init__(self, params, experiments):
        """
        Initialise the algorithm.

        :param params: The input parameters
        :param experiments: The list of experiments
        """
        from libtbx.phil import parse

        from dials.algorithms.background.simple import BackgroundAlgorithm

        # Create some default parameters
        if params is None:
            params = self.phil().fetch(parse("")).extract()
        else:
            params = params.integration.background.simple

        # Create some keyword parameters
        kwargs = {
            "model": params.model.algorithm,
            "outlier": params.outlier.algorithm,
            "min_pixels": params.min_pixels,
        }

        # Create all the keyword parameters
        if params.outlier.algorithm == "null":
            pass
        elif params.outlier.algorithm == "truncated":
            kwargs["lower"] = params.outlier.truncated.lower
            kwargs["upper"] = params.outlier.truncated.upper
        elif params.outlier.algorithm == "nsigma":
            kwargs["lower"] = params.outlier.nsigma.lower
            kwargs["upper"] = params.outlier.nsigma.upper
        elif params.outlier.algorithm == "normal":
            kwargs["min_pixels"] = params.outlier.normal.min_pixels
        elif params.outlier.algorithm == "plane":
            kwargs["fraction"] = params.outlier.plane.fraction
            kwargs["n_sigma"] = params.outlier.plane.n_sigma
        elif params.outlier.algorithm == "tukey":
            kwargs["lower"] = params.outlier.tukey.lower
            kwargs["upper"] = params.outlier.tukey.upper

        # Create the algorithm
        self._algorithm = BackgroundAlgorithm(experiments, **kwargs)

    def compute_background(self, reflections, image_volume=None):
        """
        Compute the background.

        :param reflections: The list of reflections
        """
        return self._algorithm.compute_background(
            reflections, image_volume=image_volume
        )



"""
Examine the distribution of diffraction spot intensities.

This module defines a class IntensityDist, with several methods for exploring
the distribution of measured spot intensities in an X-ray diffraction
experiment.  The user may wish to use this information to inform decisions
regarding the error model employed in analysing the data.  Data are passed in
as an unmerged MTZ file (see http://www.ccp4.ac.uk/html/mtzformat.html) and the
resulting IntensityDist instance contains the pertinent columns of data, along
with normal order statistic medians of the z-scores of the intensities, for
constructing a normal probability plot (See
https://www.itl.nist.gov/div898/handbook/eda/section3/normprpl.htm).
"""

from __future__ import annotations

import logging

import scipy.stats

from cctbx import miller
from dxtbx.model import ExperimentList

from dials.array_family import flex

log = logging.getLogger("dials.util.intensity_explorer")


class IntensityDist:
    def __init__(
        self,
        rtable,
        elist,
        calculate_variances=False,
        keep_singles=False,
        uncertainty="sigma",
        outfile=None,
    ):
        """
        Generate z-scores and a normal probability plot from a DIALS
        reflection_table and a dxtbx ExperimentList, containing the observations
        and the corresponding experiments, respectively.

        :param rtable: A reflection table object, containing at least the columns
          * ``miller_index``
          * ``intensity.sum.value``
          * ``intensity.sum.variance``
          * ``xyzobs.px.value``
        :type rtable: dials.array_family_flex_ext.reflection_table
        :param elist: A corresponding experiment list.
        :type elist: dxtbx.model.ExperimentList
        :param calculate_variances: Choose whether to calculate weighted
        aggregate variances.  Doing so incurs a performance penalty.
        Defaults to False.
        :type calculate_variances: bool
        :param keep_singles: Choose whether to keep multiplicity-1 reflections.
        Defaults to False.
        :type keep_singles: bool
        :param uncertainty: Measure of spread to use in normalising the
        z-scores, i.e. z = (I - <I>) / uncertainty.
        Possible values for uncertainty:
        * 'sigma':    Use measured sigma values;
        * 'stddev':   Use sample standard deviations calculated as
                      square-root of unbiased weighted sample variances
                      of symmetry-equivalent reflection intensities;
        Defaults to 'sigma'.
        :type uncertainty: str
        :param outfile: Filename root for output PNG plots.
        Defaults to None.
        :type: outfile: str
        """

        if not isinstance(rtable, flex.reflection_table) or not isinstance(
            elist, ExperimentList
        ):
            raise TypeError(
                "Must be called with a reflection table and an experiment list."
            )

        rtable = rtable.copy()
        # Discard unindexed reflections (only necessary because of
        # https://github.com/dials/dials/issues/615 —
        # TODO remove the line below when issue #615 is fixed).
        rtable.del_selected(rtable["id"] == -1)
        rtable["miller_index.asu"] = rtable["miller_index"]

        # Divide reflections by the space group to which they have been indexed.
        self.rtables = {
            expt.crystal.get_space_group().make_tidy(): flex.reflection_table()
            for expt in elist
        }
        for expt, sel in rtable.iterate_experiments_and_indices(elist):
            sg = expt.crystal.get_space_group().make_tidy()
            self.rtables[sg].extend(rtable.select(sel))
        # Map Miller indices to asymmetric unit.
        for space_group, rtable in self.rtables.items():
            # TODO Handle anomalous flag sensibly.  Currently assumes not anomalous.
            miller.map_to_asu(space_group.type(), False, rtable["miller_index.asu"])

        # Calculate normal probability plot data.
        self._multiplicity_mean_error_stddev(
            calculate_variances=calculate_variances, keep_singles=keep_singles
        )
        self._make_z(uncertainty)
        self._probplot_data()

        self.rtable = flex.reflection_table()
        for rtable in self.rtables.values():
            self.rtable.extend(rtable)

        if not outfile:
            outfile = ""
        self.outfile = outfile

    def _multiplicity_mean_error_stddev(
        self, calculate_variances=False, keep_singles=False
    ):
        """
        Calculate aggregate properties of grouped symmetry-equivalent reflections.

        Populate the reflection table of observations with the following
        properties:
          * ``multiplicity`` — Multiplicity of observations of a given reflection
          in the asymmetric unit;
          :type: `dials.array_family_flex_ext.int` array
          * ``intensity.mean.value`` — Mean of symmetry-equivalent reflections,
          weighted by measurement error;
          :type: `dials.array_family_flex_ext.double` array
          * ``intensity.mean.std_error`` — Standard error on the weighted mean;
          :type: `dials.array_family_flex_ext.double` array
          * (optional) ``intensity.mean.variance`` — variance of
          symmetry-equivalent reflections, weighted by measurement error;
          :type: `dials.array_family_flex_ext.double` array

        :param calculate_variances: Elect whether to calculate the weighted
        variances.  Defaults to False, to spare an expensive computation.
        :type calculate_variances: bool
        :param keep_singles: Choose whether to keep single-multiplicity
        reflections.
        :type keep_singles: bool
        """

        for key, rtable in self.rtables.items():
            # Sort the reflection table for speedier iteration.
            rtable.sort("miller_index.asu")
            # Record the positions of any multiplicity-1 reflections.
            if not keep_singles:
                singles = flex.size_t()
            # Record the multiplicities.
            multiplicity = flex.int()
            # For weighted averaging.
            weights = 1 / rtable["intensity.sum.variance"]
            sum_weights = flex.double()
            if calculate_variances:
                sum_square_weights = flex.double()
            # Calculate the weighted mean intensities.
            i_means = flex.double()
            # Calculate the standard deviations from unbiased weighted variances.
            variances = flex.double()

            # Iterate over the reflections, grouping by equivalent Miller index,
            # to calculate multiplicities, weighted mean intensities, etc..
            # Some time can be saved by only calculating variances if necessary.
            # Initial values:
            prev_index = None
            count = 1
            # The following will be set during loop iteration
            i_sum, sum_weight, sum_square_weight = None, None, None
            # One big loop through the entire reflection table:
            for j in range(rtable.size()):
                index = rtable["miller_index.asu"][j]
                weight = weights[j]
                # Aggregate within a symmetry-equivalent group of reflections:
                if index == prev_index:
                    count += 1
                    i_sum += weight * rtable["intensity.sum.value"][j]
                    sum_weight += weight
                    if calculate_variances:
                        sum_square_weight += weight * weight
                # Record the aggregated values for the group:
                elif prev_index:
                    if count == 1 and not keep_singles:
                        singles.append(j - 1)
                    multiplicity.extend(flex.int(count, count))
                    i_means.extend(flex.double(count, i_sum / sum_weight))
                    sum_weights.extend(flex.double(count, sum_weight))
                    if calculate_variances:
                        sum_square_weights.extend(flex.double(count, sum_square_weight))
                    # And reinitialise:
                    prev_index = index
                    count = 1
                    i_sum = weight * rtable["intensity.sum.value"][j]
                    sum_weight = weight
                    if calculate_variances:
                        sum_square_weight = weight * weight
                # Handle the first row:
                else:
                    prev_index = rtable["miller_index.asu"][j]
                    i_sum = weight * rtable["intensity.sum.value"][j]
                    sum_weight = weight
                    if calculate_variances:
                        sum_square_weight = weight * weight
            # Record the aggregated values for the last group:
            if count == 1 and not keep_singles:
                singles.append(rtable.size() - 1)
            multiplicity.extend(flex.int(count, count))
            i_means.extend(flex.double(count, i_sum / sum_weight))
            sum_weights.extend(flex.double(count, sum_weight))
            if calculate_variances:
                sum_square_weights.extend(flex.double(count, sum_square_weight))

            # Discard singletons:
            if not keep_singles:
                singles_del = flex.bool(rtable.size(), True)
                singles_del.set_selected(singles, False)
                multiplicity, weights, sum_weights, i_means = [
                    a.select(singles_del)
                    for a in (multiplicity, weights, sum_weights, i_means)
                ]
                rtable.del_selected(singles)
                if calculate_variances:
                    sum_square_weights = sum_square_weights.select(singles_del)

            # Record the multiplicities in the reflection table.
            rtable["multiplicity"] = multiplicity
            # Record the weighted mean intensities in the reflection table.
            rtable["intensity.mean.value"] = i_means
            # Record the standard errors on the means in the reflection table.
            rtable["intensity.mean.std_error"] = flex.sqrt(1 / sum_weights)

            if calculate_variances:
                # Initialise values:
                prev_index = None
                weighted_sum_square_residual = None
                for j in range(rtable.size()):
                    index = rtable["miller_index.asu"][j]
                    weight = weights[j]
                    residual = rtable["intensity.sum.value"][j] - i_means[j]
                    # Aggregate within a symmetry-equivalent group of reflections:
                    if index == prev_index:
                        count += 1
                        weighted_sum_square_residual += weight * residual * residual
                    # Record the aggregated value for the group:
                    elif prev_index:
                        # The weighted variance is undefined for multiplicity=1,
                        # use the measured variance instead in this case.
                        if count == 1:
                            variances.append(rtable["intensity.sum.variance"][j - 1])
                        else:
                            sum_weight = sum_weights[j - 1]
                            var_weight = 1 / (
                                sum_weight - sum_square_weights[j - 1] / sum_weight
                            )
                            variances.extend(
                                flex.double(
                                    count, weighted_sum_square_residual * var_weight
                                )
                            )
                        # Reinitialise:
                        prev_index = index
                        count = 1
                        weighted_sum_square_residual = weight * residual * residual
                    # Handle the first row:
                    else:
                        prev_index = rtable["miller_index.asu"][j]
                        count = 1
                        weighted_sum_square_residual = weight * residual * residual
                # Record the aggregated values for the last group:
                # The weighted variance is undefined for multiplicity=1,
                # use the measured variance instead in this case.
                if count == 1:
                    variances.append(rtable["intensity.sum.variance"][-1])
                else:
                    sum_weight = sum_weights[-1]
                    var_weight = 1 / (sum_weight - sum_square_weights[-1] / sum_weight)
                    variances.extend(
                        flex.double(count, weighted_sum_square_residual * var_weight)
                    )
                # Record the variances in the reflection table.
                rtable["intensity.mean.variance"] = variances

            self.rtables[key] = rtable

    def _make_z(self, uncertainty="sigma"):
        """
        Generate reflection z-scores.

        Calculate z-scores from reflection intensities, weighted mean
        intensities and a chosen measure of uncertainty in the intensity
        measurement.

        :param uncertainty: Chosen measure of uncertainty.  Options are
          * ``stddev`` — standard deviation, as calculated from the unbiased
          weighted variance aggregated amongst all symmetry-equivalent reflections;
          * ``sigma`` — measurement error for individual reflections.
        :type uncertainty: str
        """

        uncertainty_name = {
            "sigma": "intensity.sum.variance",
            "stddev": "intensity.mean.variance",
        }[uncertainty]

        for key, rtable in self.rtables.items():
            try:
                uncertainty_value = flex.sqrt(rtable[uncertainty_name])
            except KeyError:
                uncertainty_value = flex.sqrt(rtable["intensity.sum.variance"])
                log.warn(
                    """Weighted variances haven't been calculated,
      be sure to specify calculate_variances=True to use them.
      Defaulting to measured σ values as a measure of uncertainty instead."""
                )

            z = (
                rtable["intensity.sum.value"] - rtable["intensity.mean.value"]
            ) / uncertainty_value
            rtable["intensity.z_score"] = z

            self.rtables[key] = rtable

    def _probplot_data(self):
        """Generate the data for a normal probability plot of z-scores."""

        for key, rtable in self.rtables.items():
            order = flex.sort_permutation(rtable["intensity.z_score"])
            osm = flex.double(rtable.size(), 0)
            probplot = scipy.stats.probplot(rtable["intensity.z_score"], fit=False)
            osm.set_selected(order, flex.double(probplot[0]))
            rtable["intensity.order_statistic_medians"] = osm

            self.rtables[key] = rtable



from __future__ import annotations

from scitbx import matrix

from dials.algorithms.refinement.parameterisation.model_parameters import (
    ModelParameterisation,
    Parameter,
)
from dials.algorithms.refinement.refinement_helpers import dR_from_axis_and_angle


class BeamMixin:
    """Mix-in class defining some functionality unique to beam parameterisations
    that can be shared by static and scan-varying versions"""

    @staticmethod
    def _build_p_list(s0, goniometer, parameter_type=Parameter):
        """Build the list of parameters, using the parameter_type callback to
        select between versions of the Parameter class"""

        # Set up the parameters
        if goniometer:
            spindle = matrix.col(goniometer.get_rotation_axis())
            s0_plane_dir2 = s0.cross(spindle).normalize()
            s0_plane_dir1 = s0_plane_dir2.cross(s0).normalize()
        else:
            s0_plane_dir1 = s0.ortho().normalize()
            s0_plane_dir2 = s0.cross(s0_plane_dir1).normalize()

        # rotation around s0_plane_dir1
        mu1 = parameter_type(0.0, s0_plane_dir1, "angle (mrad)", "Mu1")
        # rotation around s0_plane_dir2
        mu2 = parameter_type(0.0, s0_plane_dir2, "angle (mrad)", "Mu2")
        # length of s0
        nu = parameter_type(
            s0.length(), axis=None, ptype="wavenumber (Angstroem^-1)", name="nu"
        )

        # build the parameter list in a specific,  maintained order
        p_list = [mu1, mu2, nu]

        return p_list

    @staticmethod
    def _compose_core(is0, ipn, mu1, mu2, nu, mu1_axis, mu2_axis):

        # convert angles to radians
        mu1rad, mu2rad = mu1 / 1000.0, mu2 / 1000.0

        # compose rotation matrices and their first order derivatives
        Mu1 = (mu1_axis).axis_and_angle_as_r3_rotation_matrix(mu1rad, deg=False)
        dMu1_dmu1 = dR_from_axis_and_angle(mu1_axis, mu1rad, deg=False)

        Mu2 = (mu2_axis).axis_and_angle_as_r3_rotation_matrix(mu2rad, deg=False)
        dMu2_dmu2 = dR_from_axis_and_angle(mu2_axis, mu2rad, deg=False)

        # compose new state
        Mu21 = Mu2 * Mu1
        s0_new_dir = (Mu21 * is0).normalize()
        pn_new_dir = (Mu21 * ipn).normalize()
        s0 = nu * s0_new_dir

        # calculate derivatives of the beam direction wrt angles:
        #  1) derivative wrt mu1
        dMu21_dmu1 = Mu2 * dMu1_dmu1
        ds0_new_dir_dmu1 = dMu21_dmu1 * is0

        #  2) derivative wrt mu2
        dMu21_dmu2 = dMu2_dmu2 * Mu1
        ds0_new_dir_dmu2 = dMu21_dmu2 * is0

        # calculate derivatives of the attached beam vector, converting
        # parameters back to mrad
        ds0_dval = [
            ds0_new_dir_dmu1 * nu / 1000.0,
            ds0_new_dir_dmu2 * nu / 1000.0,
            s0_new_dir,
        ]

        return (s0, pn_new_dir), ds0_dval


class BeamParameterisation(ModelParameterisation, BeamMixin):
    """A parameterisation of a Beam model.

    The Beam direction and energy are parameterised using angles expressed in
    mrad and wavenumber in inverse Angstroms. A goniometer can be provided (if
    present in the experiment) to ensure a consistent definition of the beam
    rotation angles with respect to the spindle-beam plane."""

    def __init__(self, beam, goniometer=None, experiment_ids=None):
        """Initialise the BeamParameterisation object

        Args:
            beam: A dxtbx Beam object to be parameterised.
            goniometer: An optional dxtbx Goniometer object. Defaults to None.
            experiment_ids (list): The experiment IDs affected by this
                parameterisation. Defaults to None, which is replaced by [0].
        """
        # The state of the beam model consists of the s0 vector that it is
        # modelling. The initial state is the direction of this vector at the point
        # of initialisation, plus the direction of the orthogonal polarization
        # normal vector. Future states are composed by rotations around axes
        # perpendicular to that direction and normalisation specified by the
        # wavenumber (inverse wavelength).

        # Set up the initial state
        if experiment_ids is None:
            experiment_ids = [0]
        s0 = matrix.col(beam.get_s0())
        istate = {
            "unit_s0": matrix.col(beam.get_unit_s0()),
            "polarization_normal": matrix.col(beam.get_polarization_normal()),
        }

        # build the parameter list
        p_list = self._build_p_list(s0, goniometer)

        # set up the base class
        ModelParameterisation.__init__(
            self, beam, istate, p_list, experiment_ids=experiment_ids
        )

        # call compose to calculate all the derivatives
        self.compose()

        return

    def compose(self):

        # extract direction from the initial state
        ius0 = self._initial_state["unit_s0"]
        ipn = self._initial_state["polarization_normal"]

        # extract parameters from the internal list
        mu1, mu2, nu = self._param

        # calculate new s0 and derivatives
        (s0, pn), self._dstate_dp = self._compose_core(
            ius0,
            ipn,
            mu1.value,
            mu2.value,
            nu.value,
            mu1_axis=mu1.axis,
            mu2_axis=mu2.axis,
        )

        # now update the model with its new s0 and polarization vector
        self._model.set_s0(s0)
        self._model.set_polarization_normal(pn)

        return

    def get_state(self):

        # only a single beam exists, so no multi_state_elt argument is allowed
        return matrix.col(self._model.get_s0())



# LIBTBX_SET_DISPATCHER_NAME dials.filter_reflections


from __future__ import annotations

import logging
import token
from operator import itemgetter
from tokenize import TokenError, generate_tokens, untokenize

from cctbx import uctbx
from libtbx.phil import parse

from dials.algorithms.integration import filtering
from dials.array_family import flex
from dials.util import Sorry, log, show_mail_handle_errors, tabulate
from dials.util.filter_reflections import SumAndPrfIntensityReducer, SumIntensityReducer
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

logger = logging.getLogger("dials")

help_message = """

This program takes reflection files as input and filters them based on user-
specified criteria, to write out a subset of the original file.

Filtering is first done by evaluating the optional boolean 'flag_expression'
using reflection flag values. The operators allowed are '&' for 'and', '|' for
'or', and '~' for 'not'. Expressions may contain nested sub-expressions using
parentheses.

Following this, optional additional filters are applied according to values in
the reflection table, such as by resolution or user-defined masks.

If a reflection file is passed in to the program but no filtering parameters
are set, a table will be printed, giving the flag values present in the
reflection file.


Examples::

  dials.filter_reflections refined.refl \
    flag_expression=used_in_refinement

  dials.filter_reflections integrated.refl \
    flag_expression="integrated & ~reference_spot"

  dials.filter_reflections integrated.refl \
    flag_expression="indexed & (failed_during_summation | failed_during_profile_fitting)"

  dials.filter_reflections indexed.refl indexed.expt \
    d_max=20 d_min=2.5
"""

phil_str = """
    output {
    reflections = 'filtered.refl'
        .type = str
        .help = "The filtered reflections output filename"
    }

    flag_expression = None
    .type = str
    .help = "Boolean expression to select reflections based on flag values"

    id = None
    .type = ints(value_min=0)
    .help = "Select reflections by experiment IDs"

    panel = None
    .type = ints(value_min=0)
    .help = "Select reflections by panels they intersect"

    d_min = None
    .type = float
    .help = "The maximum resolution"

    d_max = None
    .type = float
    .help = "The minimum resolution"

    partiality {
    min = None
        .type = float(value_min = 0, value_max = 1)
        .help = "The minimum reflection partiality for inclusion."
    max = None
        .type = float(value_min = 0, value_max = 1)
        .help = "The maximum reflection partiality for inclusion."
    }

    select_good_intensities = False
    .type = bool
    .help = "Combined filter to select only fully integrated and"
            "trustworthy intensities"

    remove_by_index = None
      .type = ints(value_min=0)
      .help = "Remove reflections by row index in the table"
      .expert_level = 6

    dead_time
    {
    value = 0
        .help = "Detector dead time in ms, assumed to be at the end of the"
                "exposure time."
        .type = float(value_min=0)

    reject_fraction = 0
        .help = "Reject reflections which overlap by more than the given"
                "fraction with the dead region of the image."
        .type = float(value_min=0, value_max=1)
    }

    include scope dials.util.masking.phil_scope ice_rings
"""

phil_scope = parse(phil_str, process_includes=True)


def eval_flag_expression(expression, reflections):
    """Test a Boolean expression of reflection flags for validity then
    evaluate it"""

    result = []
    tokens = iter(expression.split("\n"))
    g = generate_tokens(lambda: next(tokens))

    flags = list(flex.reflection_table.flags.names.items())
    flags.sort(key=itemgetter(0))
    flag_names, _ = zip(*flags)

    # define shorthand function
    def get_flag(flag):
        return reflections.get_flags(getattr(reflections.flags, flag))

    while True:
        # Extract next token, catching unmatched brackets
        try:
            toknum, tokval, _, _, _ = next(g)
        except TokenError:
            raise Sorry(f"errors found in {expression}")
        except StopIteration:
            break

        # Skip newline characters
        if toknum == token.NEWLINE:
            continue

        # Catch unwanted token types
        if toknum not in [token.OP, token.NAME, token.ENDMARKER]:
            raise Sorry(f"invalid token {token.tok_name[toknum]} found in {expression}")

        # Catch unwanted operators
        if toknum is token.OP and tokval not in "()|&~":
            raise Sorry(f"unrecognised operators found in {expression}")

        # Catch unrecognised flag names
        if toknum is token.NAME and tokval not in flag_names:
            raise Sorry(f"unrecognised flag name: {tokval}")

        # Replace names with valid lookups in the reflection table
        if toknum is token.NAME:
            # ("NAME", "get_flags")
            # ("OP", "(")
            # ("STRING", "'indexed'")
            # ("OP", ")")
            result.extend(
                [
                    (token.NAME, "get_flag"),
                    (token.OP, "("),
                    (token.STRING, repr(tokval)),
                    (token.OP, ")"),
                ]
            )
        else:
            result.append((toknum, tokval))

    # Evaluate and return the result
    return eval(untokenize(result), {"get_flag": get_flag, "__builtins__": None}, {})


def run_analysis(flags, reflections):
    """Print a table of flags present in the reflections file"""

    header = ["flag", "nref"]
    rows = []
    for name, val in flags:
        n = (reflections.get_flags(val)).count(True)
        if n > 0:
            rows.append([name, "%d" % n])
    if rows:
        print(tabulate(rows, header))
    else:
        print("No flags set")


def run_filtering(params, experiments, reflections):
    """Execute the script."""

    # Check params
    if params.d_min is not None and params.d_max is not None:
        if params.d_min > params.d_max:
            raise Sorry("d_min must be less than d_max")
    if params.d_min is not None or params.d_max is not None or params.ice_rings.filter:
        if "d" not in reflections:
            if experiments and any(experiments.crystals()):
                # Calculate d-spacings from the miller indices
                print(
                    "Reflection table does not have resolution information. "
                    "Attempting to calculate this from the experiment list"
                )
                sel = reflections["id"] >= 0
                if sel.count(False) > 0:
                    print(
                        "Removing {} reflections with negative experiment id".format(
                            sel.count(False)
                        )
                    )
                reflections = reflections.select(sel)
                reflections.compute_d(experiments)
            elif experiments:
                # Calculate d-spacings from the observed reflection centroids
                if "rlp" not in reflections:
                    if "xyzobs.mm.value" not in reflections:
                        reflections.centroid_px_to_mm(experiments)
                    reflections.map_centroids_to_reciprocal_space(experiments)
                d_star_sq = flex.pow2(reflections["rlp"].norms())
                reflections["d"] = uctbx.d_star_sq_as_d(d_star_sq)
            else:
                raise Sorry(
                    "reflection table has no resolution information "
                    "and no experiment list provided to calculate it"
                )

    # Check params
    if params.partiality.min is not None and params.partiality.max is not None:
        if params.min > params.max:
            raise Sorry("partiality.min must be less than partiality.d_max")
    if params.partiality.min is not None or params.partiality.max is not None:
        if "partiality" not in reflections:
            raise Sorry("Reflection table has no partiality information")

    print(f"{len(reflections)} reflections loaded")

    # Filter by logical expression using flags
    if params.flag_expression is not None:
        inc = eval_flag_expression(params.flag_expression, reflections)
        reflections = reflections.select(inc)

    print(f"Selected {len(reflections)} reflections by flags")

    # Filter based on experiment ID
    if params.id:
        selection = reflections["id"] == params.id[0]
        for exp_id in params.id[1:]:
            selection = selection | (reflections["id"] == exp_id)
        reflections = reflections.select(selection)
        print(f"Selected {len(reflections)} reflections by experiment id")

    # Filter based on panel number
    if params.panel:
        selection = reflections["panel"] == params.panel[0]
        for pnl_id in params.panel[1:]:
            selection = selection | (reflections["panel"] == pnl_id)
        reflections = reflections.select(selection)
        print(f"Selected {len(reflections)} reflections by panel number")

    # Filter based on resolution
    if params.d_min is not None:
        selection = reflections["d"] >= params.d_min
        reflections = reflections.select(selection)
        print(f"Selected {len(reflections)} reflections with d >= {params.d_min:f}")

    # Filter based on resolution
    if params.d_max is not None:
        selection = reflections["d"] <= params.d_max
        reflections = reflections.select(selection)
        print(f"Selected {len(reflections)} reflections with d <= {params.d_max:f}")

    # Filter based on partiality
    if params.partiality.min is not None:
        selection = reflections["partiality"] >= params.partiality.min
        reflections = reflections.select(selection)
        print(
            "Selected %d reflections with partiality >= %f"
            % (len(reflections), params.partiality.min)
        )

    # Filter based on partiality
    if params.partiality.max is not None:
        selection = reflections["partiality"] <= params.partiality.max
        reflections = reflections.select(selection)
        print(
            "Selected %d reflections with partiality <= %f"
            % (len(reflections), params.partiality.max)
        )

    # 'Good' intensity selection
    if params.select_good_intensities:
        if "intensity.sum.variance" not in reflections:
            raise Sorry("No intensity.sum.variance in reflection table.")
        if "intensity.prf.variance" in reflections:
            reducer = SumAndPrfIntensityReducer
        else:
            reducer = SumIntensityReducer
        try:
            reflections = reducer.filter_bad_variances(reflections)
            reflections = reducer.combine_and_filter_partials(
                reflections, partiality_threshold=0.99, combine_partials=False
            )
        except ValueError as e:
            raise Sorry(e)

    if params.remove_by_index:
        to_remove = flex.size_t(params.remove_by_index)
        filter = flex.bool(len(reflections), True)
        filter.set_selected(to_remove, False)
        reflections = reflections.select(filter)
        print("Selected %d reflections after removing by index" % (len(reflections)))

    # Dead time filter
    if params.dead_time.value > 0:
        reflections = filter_by_dead_time(
            reflections,
            experiments,
            params.dead_time.value,
            params.dead_time.reject_fraction,
        )

    # Filter powder rings
    if params.ice_rings.filter:
        d_min = params.ice_rings.d_min
        width = params.ice_rings.width

        if d_min is None:
            d_min = flex.min(reflections["d"])

        ice_filter = filtering.PowderRingFilter(
            params.ice_rings.unit_cell,
            params.ice_rings.space_group.group(),
            d_min,
            width,
        )

        ice_sel = ice_filter(reflections["d"])

        print("Rejecting %i reflections at ice ring resolution" % ice_sel.count(True))
        reflections = reflections.select(~ice_sel)

    # Save filtered reflections to file
    if params.output.reflections:
        print(f"Saving {len(reflections)} reflections to {params.output.reflections}")
        reflections.as_file(params.output.reflections)


def filter_by_dead_time(reflections, experiments, dead_time=0, reject_fraction=0):
    if not experiments:
        raise Sorry("an experiment list must be provided to filter by dead time")

    if len(experiments) > 1:
        raise Sorry("only a single experiment is supported in this mode")
    experiment = experiments[0]

    sel = reflections.get_flags(reflections.flags.integrated)
    reflections = reflections.select(sel)

    if not reflections:
        raise Sorry("no integrated reflections present")

    goniometer = experiment.goniometer
    beam = experiment.beam

    phi_range = reflections.compute_phi_range(
        goniometer.get_rotation_axis(),
        beam.get_s0(),
        experiment.profile.sigma_m(deg=False),
        experiment.profile.n_sigma(),
    )
    phi1, phi2 = phi_range.parts()

    scan = experiment.scan
    exposure_time = scan.get_exposure_times()[0]
    assert scan.get_exposure_times().all_eq(exposure_time)
    phi_start, phi_width = scan.get_oscillation(deg=False)
    phi_range_dead = phi_width * (dead_time / 1000) / exposure_time

    sel_good = flex.bool(len(reflections), True)

    start, end = scan.get_array_range()
    for i in range(start, end):
        phi_dead_start = phi_start + (i + 1) * phi_width - phi_range_dead
        phi_dead_end = phi_dead_start + phi_range_dead

        left = phi1.deep_copy()
        left.set_selected(left < phi_dead_start, phi_dead_start)

        right = phi2.deep_copy()
        right.set_selected(right > phi_dead_end, phi_dead_end)

        overlap = (right - left) / (phi2 - phi1)

        sel = overlap > reject_fraction

        sel_good.set_selected(sel, False)
        print("Rejecting %i reflections from image %i" % (sel.count(True), i))

    print(
        "Keeping %i reflections (rejected %i)"
        % (sel_good.count(True), sel_good.count(False))
    )

    return reflections.select(sel_good)


@show_mail_handle_errors()
def run(args=None):
    """Run the command line filtering script."""

    flags = list(flex.reflection_table.flags.names.items())
    flags.sort(key=itemgetter(0))

    # The script usage
    usage = "usage: dials.filter_reflections [options] experiment.expt"

    # Create the parser
    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        epilog=help_message,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    log.config(verbosity=options.verbose)

    if not reflections:
        parser.print_help()
        return
    if len(reflections) != 1:
        raise Sorry("Exactly 1 reflection file must be specified")
    reflections = reflections[0]

    # Check if any filter has been set using diff_phil
    filter_def = [
        o for o in parser.diff_phil.objects if o.name not in ["input", "output"]
    ]
    if not filter_def:
        print("No filter specified. Performing analysis instead.")
        run_analysis(flags, reflections)
    else:
        run_filtering(params, experiments, reflections)


if __name__ == "__main__":
    run()



from __future__ import annotations

import logging

import scitbx.matrix
from cctbx import crystal, sgtbx
from cctbx.crystal_orientation import crystal_orientation
from cctbx.sgtbx import change_of_basis_op, subgroups
from cctbx.sgtbx.bravais_types import bravais_lattice
from dxtbx.model import Crystal
from rstbx.dps_core.lepage import iotbx_converter
from scitbx.array_family import flex

logger = logging.getLogger(__name__)


def metric_supergroup(group):
    return (
        sgtbx.space_group_info(group=group)
        .type()
        .expand_addl_generators_of_euclidean_normalizer(True, True)
        .build_derived_acentric_group()
    )


def groups_cache(fn):
    class MultiClassCache(object):

        "A set of caches for different bravais types"

        instances = {}

        def __new__(cls, classname):
            if classname not in cls.instances:
                cls.instances[classname] = {}
            return cls.instances[classname]

    def wrapped_calc(group_info: sgtbx.space_group_info, bravais_t: str):
        cache = MultiClassCache(bravais_t)
        info_str = group_info.type().lookup_symbol()
        try:
            result = cache[info_str]
        except KeyError:
            result = fn(group_info, bravais_t)
            cache[info_str] = result
        return result

    return wrapped_calc


@groups_cache
def calc_acentric_subgroups(
    lattice_group_info: sgtbx.space_group_info, target_bravais_str: str
):
    # Get list of sub-spacegroups
    subgrs = subgroups.subgroups(lattice_group_info).groups_parent_setting()

    # Order sub-groups
    sort_values = flex.double()
    for group in subgrs:
        order_z = group.order_z()
        space_group_number = sgtbx.space_group_type(group, False).number()
        assert 1 <= space_group_number <= 230
        sort_values.append(order_z * 1000 + space_group_number)
    perm = flex.sort_permutation(sort_values, reverse=True)
    acentric_subgroups = []
    acentric_supergroups = []
    cb_ops = []
    for i_perm in perm:
        acentric_subgroup = subgrs[i_perm]
        cb_op = acentric_subgroup.info().type().cb_op()
        ref_acentric_subgroup = acentric_subgroup.change_basis(cb_op)
        # Ignore unwanted groups
        if str(bravais_lattice(group=ref_acentric_subgroup)) != target_bravais_str:
            continue
        acentric_subgroups.append(acentric_subgroup)
        cb_ops.append(cb_op)
        acentric_supergroups.append(metric_supergroup(acentric_subgroup))
    return acentric_subgroups, acentric_supergroups, cb_ops


def find_matching_symmetry(
    unit_cell,
    target_space_group,
    max_delta=5,
    best_monoclinic_beta=True,
    target_bravais_str=None,
):
    cs = crystal.symmetry(unit_cell=unit_cell, space_group=sgtbx.space_group())
    if target_bravais_str is None:
        target_bravais_str = str(
            bravais_lattice(group=target_space_group.info().reference_setting().group())
        )
    best_subgroup = None
    best_angular_difference = 1e8

    # code based on cctbx/sgtbx/lattice_symmetry.py but optimised to only
    # look at subgroups with the correct bravais type

    input_symmetry = cs
    # Get cell reduction operator
    cb_op_inp_minimum = input_symmetry.change_of_basis_op_to_minimum_cell()

    # New symmetry object with changed basis
    minimum_symmetry = input_symmetry.change_basis(cb_op_inp_minimum)

    # Get highest symmetry compatible with lattice
    lattice_group = sgtbx.lattice_symmetry_group(
        minimum_symmetry.unit_cell(),
        max_delta=max_delta,
        enforce_max_delta_for_generated_two_folds=True,
    )

    lattice_group_info = lattice_group.info()
    acentric_subgroups, acentric_supergroups, cb_ops = calc_acentric_subgroups(
        lattice_group_info,
        target_bravais_str,
    )

    for acentric_subgroup, acentric_supergroup, cb_op_minimum_ref in zip(
        acentric_subgroups, acentric_supergroups, cb_ops
    ):

        # Make symmetry object: unit-cell + space-group
        # The unit cell is potentially modified to be exactly compatible
        # with the space group symmetry.
        subsym = crystal.symmetry(
            unit_cell=minimum_symmetry.unit_cell(),
            space_group=acentric_subgroup,
            assert_is_compatible_unit_cell=False,
        )
        # Convert subgroup to reference setting
        ref_subsym = subsym.change_basis(cb_op_minimum_ref)

        # Choose best setting for monoclinic and orthorhombic systems
        cb_op_best_cell = ref_subsym.change_of_basis_op_to_best_cell(
            best_monoclinic_beta=best_monoclinic_beta
        )

        best_subsym = ref_subsym.change_basis(cb_op_best_cell)
        # Total basis transformation
        cb_op_best_cell = change_of_basis_op(
            str(cb_op_best_cell), stop_chars="", r_den=144, t_den=144
        )
        cb_op_minimum_ref = change_of_basis_op(
            str(cb_op_minimum_ref), stop_chars="", r_den=144, t_den=144
        )
        cb_op_inp_minimum = change_of_basis_op(
            str(cb_op_inp_minimum), stop_chars="", r_den=144, t_den=144
        )
        cb_op_inp_best = cb_op_best_cell * cb_op_minimum_ref * cb_op_inp_minimum
        # Use identity change-of-basis operator if possible
        if best_subsym.unit_cell().is_similar_to(input_symmetry.unit_cell()):
            cb_op_corr = cb_op_inp_best.inverse()
            try:
                best_subsym_corr = best_subsym.change_basis(cb_op_corr)
            except RuntimeError as e:
                if str(e).find("Unsuitable value for rational rotation matrix.") < 0:
                    raise
            else:
                if best_subsym_corr.space_group() == best_subsym.space_group():
                    cb_op_inp_best = cb_op_corr * cb_op_inp_best

        max_angular_difference = sgtbx.lattice_symmetry_find_max_delta(
            reduced_cell=minimum_symmetry.unit_cell(), space_group=acentric_supergroup
        )

        if max_angular_difference < best_angular_difference:
            best_angular_difference = max_angular_difference
            best_subgroup = {
                "subsym": subsym,
                "ref_subsym": ref_subsym,
                "best_subsym": best_subsym,
                "cb_op_inp_best": cb_op_inp_best,
                "max_angular_difference": max_angular_difference,
            }

    if best_subgroup is not None:
        return best_subgroup


class SymmetryHandler:
    def __init__(self, unit_cell=None, space_group=None, max_delta=5):

        self._max_delta = max_delta
        self.target_symmetry_primitive = None
        self.target_symmetry_reference_setting = None
        self.cb_op_inp_ref = None
        self.cb_op_inp_best = None

        target_space_group = space_group
        if target_space_group is not None:
            target_space_group = target_space_group.build_derived_patterson_group()

        if unit_cell is not None:

            assert (
                space_group
            ), "space_group must be provided in combination with unit_cell"

            self.target_symmetry_inp = crystal.symmetry(
                unit_cell=unit_cell, space_group=target_space_group
            )
            if str(bravais_lattice(group=target_space_group)) == "aP":
                self.cb_op_inp_ref = (
                    self.target_symmetry_inp.change_of_basis_op_to_minimum_cell()
                )
            else:
                self.cb_op_inp_ref = (
                    self.target_symmetry_inp.change_of_basis_op_to_reference_setting()
                )
            self.target_symmetry_reference_setting = (
                self.target_symmetry_inp.change_basis(self.cb_op_inp_ref)
            )
            self.cb_op_inp_best = (
                self.target_symmetry_reference_setting.change_of_basis_op_to_best_cell()
                * self.cb_op_inp_ref
            )

        elif target_space_group is not None:
            self.target_symmetry_inp = crystal.symmetry(space_group=target_space_group)
            self.cb_op_inp_ref = (
                target_space_group.info().change_of_basis_op_to_reference_setting()
            )
            self.target_symmetry_reference_setting = crystal.symmetry(
                space_group=target_space_group.change_basis(self.cb_op_inp_ref)
            )

        cb_op_reference_to_primitive = (
            self.target_symmetry_reference_setting.change_of_basis_op_to_primitive_setting()
        )
        if unit_cell:
            self.target_symmetry_primitive = (
                self.target_symmetry_reference_setting.change_basis(
                    cb_op_reference_to_primitive
                )
            )
        else:
            self.target_symmetry_primitive = crystal.symmetry(
                space_group=self.target_symmetry_reference_setting.space_group().change_basis(
                    cb_op_reference_to_primitive
                )
            )
        self.cb_op_ref_inp = self.cb_op_inp_ref.inverse()
        self.cb_op_primitive_inp = (
            self.cb_op_ref_inp * cb_op_reference_to_primitive.inverse()
        )

        if self.target_symmetry_reference_setting:
            logger.debug(
                "Target symmetry (reference setting):\n%s",
                self.target_symmetry_reference_setting,
            )
        if self.target_symmetry_primitive:
            logger.debug(
                "Target symmetry (primitive cell):\n%s",
                self.target_symmetry_primitive,
            )
        logger.debug("cb_op primitive->input: %s", self.cb_op_primitive_inp)

    def apply_symmetry(self, crystal_model):
        """Apply symmetry constraints to a crystal model.

        Returns the crystal model (with symmetry constraints applied) in the same
        setting as provided as input. The cb_op returned by the method is
        that necessary to transform that model to the user-provided
        target symmetry.

        Args:
            crystal_model (dxtbx.model.Crystal): The input crystal model to which to
              apply symmetry constraints.

        Returns: (dxtbx.model.Crystal, cctbx.sgtbx.change_of_basis_op):
        The crystal model with symmetry constraints applied, and the change_of_basis_op
        that transforms the returned model to the user-specified target symmetry.
        """
        if not (
            self.target_symmetry_primitive
            and self.target_symmetry_primitive.space_group()
        ):
            return crystal, sgtbx.change_of_basis_op()

        target_space_group = self.target_symmetry_primitive.space_group()

        A = crystal_model.get_A()

        max_delta = self._max_delta
        items = iotbx_converter(crystal_model.get_unit_cell(), max_delta=max_delta)
        target_sg_ref = target_space_group.info().reference_setting().group()
        best_angular_difference = 1e8

        best_subgroup = None
        for item in items:
            if bravais_lattice(group=target_sg_ref) != item["bravais"]:
                continue
            if item["max_angular_difference"] < best_angular_difference:
                best_angular_difference = item["max_angular_difference"]
                best_subgroup = item

        if best_subgroup is None:
            return None, None

        cb_op_inp_best = best_subgroup["cb_op_inp_best"]
        best_subsym = best_subgroup["best_subsym"]
        ref_subsym = best_subgroup["ref_subsym"]
        cb_op_ref_best = ref_subsym.change_of_basis_op_to_best_cell()
        cb_op_best_ref = cb_op_ref_best.inverse()
        cb_op_inp_ref = cb_op_best_ref * cb_op_inp_best
        cb_op_ref_inp = cb_op_inp_ref.inverse()

        orient = crystal_orientation(A, True)
        orient_ref = orient.change_basis(
            scitbx.matrix.sqr((cb_op_inp_ref).c().as_double_array()[0:9]).transpose()
        )
        constrain_orient = orient_ref.constrain(best_subgroup["system"])
        direct_matrix = constrain_orient.direct_matrix()

        a = scitbx.matrix.col(direct_matrix[:3])
        b = scitbx.matrix.col(direct_matrix[3:6])
        c = scitbx.matrix.col(direct_matrix[6:9])
        model = Crystal(a, b, c, space_group=target_sg_ref)
        assert target_sg_ref.is_compatible_unit_cell(model.get_unit_cell())

        model = model.change_basis(cb_op_ref_inp)

        if self.cb_op_inp_best is not None:
            # Then the unit cell has been provided: this is the cb_op to map to the
            # user-provided input unit cell
            return model, self.cb_op_inp_best.inverse() * cb_op_inp_best
        if not self.cb_op_ref_inp.is_identity_op():
            if self.target_symmetry_inp.space_group() == best_subsym.space_group():
                # Handle where e.g. the user has requested I2 instead of the reference C2
                return model, cb_op_inp_best
            # The user has specified a setting that is not the reference setting
            return model, self.cb_op_ref_inp * cb_op_inp_ref
        # Default to reference setting
        # This change of basis op will ensure that we get the best beta angle without
        # changing the centring (e.g. from C2 to I2)
        cb_op_ref_best = ref_subsym.change_of_basis_op_to_best_cell(
            best_monoclinic_beta=False
        )
        return model, cb_op_ref_best * cb_op_inp_ref



"""Test the use of BlockCalculator in ReflectionManager as used in preparation
for scan-varying refinement. This exercises the issue originally flagged in
https://github.com/dials/dials/issues/511"""


from __future__ import annotations

from math import pi

import pytest

from dxtbx.model.experiment_list import Experiment, ExperimentList
from scitbx.array_family import flex

from dials.algorithms.refinement.reflection_manager import BlockCalculator

from . import geometry_phil


def create_experiments(image_start=1):

    # Create models
    from libtbx.phil import parse

    overrides = """geometry.parameters.crystal.a.length.range = 10 50
  geometry.parameters.crystal.b.length.range = 10 50
  geometry.parameters.crystal.c.length.range = 10 50"""
    master_phil = parse(geometry_phil)

    from .setup_geometry import Extract

    models = Extract(master_phil, overrides)

    detector = models.detector
    goniometer = models.goniometer
    crystal = models.crystal
    beam = models.beam

    # Build a mock scan for a 72 degree sequence
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    scan = sf.make_scan(
        image_range=(image_start, image_start + 720 - 1),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(720)),
        deg=True,
    )

    # No matter what image_start is, scan should start at 0.0 and end at 72.0 deg
    assert scan.get_oscillation_range(deg=True) == (0.0, 72.0)

    # Create an ExperimentList
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            crystal=crystal,
            imageset=None,
        )
    )

    return experiments


def generate_reflections(experiments):

    from cctbx.sgtbx import space_group, space_group_symbols

    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansExperimentsPredictor,
        ScansRayPredictor,
    )
    from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

    detector = experiments[0].detector
    crystal = experiments[0].crystal

    # All indices in a 2.0 Angstrom sphere
    resolution = 2.0
    index_generator = IndexGenerator(
        crystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Predict rays within the sequence range
    scan = experiments[0].scan
    sequence_range = scan.get_oscillation_range(deg=False)
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs = ray_predictor(indices)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(detector, obs_refs)
    obs_refs = obs_refs.select(intersects)

    # Make a reflection predictor and re-predict for all these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs["id"] = flex.int(len(obs_refs), 0)
    obs_refs = ref_predictor(obs_refs)

    # Set 'observed' centroids from the predicted ones
    obs_refs["xyzobs.mm.value"] = obs_refs["xyzcal.mm"]

    return obs_refs


def test_per_width_and_per_image_are_equivalent():

    # Scan starting at image 1
    experiments = create_experiments(1)
    reflections = generate_reflections(experiments)

    # Check scan is consistent with the reflections
    phi_obs = reflections["xyzobs.mm.value"].parts()[2] * 180.0 / pi
    z_cal = reflections["xyzcal.px"].parts()[2]
    for phi, z in zip(phi_obs, z_cal):
        z2 = experiments[0].scan.get_array_index_from_angle(phi, deg=True)
        assert z == pytest.approx(z2)

    # Set blocks with per_width
    from copy import deepcopy

    block_calculator = BlockCalculator(experiments, deepcopy(reflections))
    im_width = experiments[0].scan.get_oscillation(deg=False)[1]
    r_pw = block_calculator.per_width(im_width, deg=False)

    # Set blocks with per_image
    block_calculator = BlockCalculator(experiments, deepcopy(reflections))
    r_pi = block_calculator.per_image()

    # Check block assignment is the same
    assert r_pw["block"].all_eq(r_pi["block"])
    for bc1, bc2 in zip(r_pw["block_centre"], r_pi["block_centre"]):
        assert bc1 == pytest.approx(bc2)

    # Scan starting at image 100
    experiments = create_experiments(100)
    reflections100 = generate_reflections(experiments)

    # Check reflections and experiments are as expected
    assert len(reflections100) == len(reflections)
    for a, b in zip(reflections.rows(), reflections100.rows()):
        assert a["xyzcal.mm"] == b["xyzcal.mm"]
    assert experiments[0].scan.get_oscillation(deg=False)[1] == im_width
    reflections = reflections100

    # Check scan is consistent with the reflections
    phi_obs = reflections["xyzobs.mm.value"].parts()[2] * 180.0 / pi
    z_cal = reflections["xyzcal.px"].parts()[2]
    for phi, z in zip(phi_obs, z_cal):
        z2 = experiments[0].scan.get_array_index_from_angle(phi, deg=True)
        assert z == pytest.approx(z2)

    # Set blocks with per_width
    block_calculator = BlockCalculator(experiments, deepcopy(reflections))
    assert experiments[0].scan.get_oscillation(deg=False)[1] == im_width
    r_pw_ = block_calculator.per_width(im_width, deg=False)

    # Block centres should have all increased by 99.0
    for a, b in zip(r_pw["block_centre"], r_pw_["block_centre"]):
        assert b == pytest.approx(a + 99.0)
    r_pw = r_pw_

    # Set blocks with per_image
    block_calculator = BlockCalculator(experiments, deepcopy(reflections))
    r_pi = block_calculator.per_image()

    # Should still give the same results as per_width
    assert r_pw["block"].all_eq(r_pi["block"])
    for bc1, bc2 in zip(r_pw["block_centre"], r_pi["block_centre"]):
        assert bc1 == pytest.approx(bc2)



from __future__ import annotations

from scitbx.array_family import flex

from dials.algorithms.refinement.parameterisation.model_parameters import (
    ModelParameterisation,
    Parameter,
)
from dials_refinement_helpers_ext import GaussianSmoother as GS

# reusable PHIL string for options affecting scan-varying parameterisation
phil_str = """
smoother
  .help = "Options that affect scan-varying parameterisation"
  .expert_level = 1
{
  interval_width_degrees = 36.0
    .help = "Width of scan between checkpoints in degrees. Can be set to Auto."
    .type = float(value_min=0.)

  absolute_num_intervals = None
    .help = "Number of intervals between checkpoints if scan_varying"
            "refinement is requested. If set, this overrides"
            "interval_width_degrees"
    .type = int(value_min=1)
}
"""


class ScanVaryingParameterSet(Parameter):
    """Testing a class for a scan-varying parameter, in which values at rotation
    angle phi may be derived using smoothed interpolation between checkpoint
    values stored here. Externally, this is presented as a set of parameters.

    num_samples is the number of checkpoints. Other arguments are as Parameter.
    """

    def __init__(
        self,
        value,
        num_samples=5,
        axis=None,
        ptype=None,
        name="ScanVaryingParameterSet",
    ):

        assert num_samples >= 2  # otherwise use scan-independent parameterisation

        value = [value] * num_samples
        self._name_stem = name
        name = [
            e + "_sample%d" % i for i, e in enumerate([self._name_stem] * num_samples)
        ]

        Parameter.__init__(self, value, axis, ptype, name)

        self._esd = [None] * num_samples
        self._num_samples = num_samples

        return

    def __len__(self):
        return self._num_samples

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        assert len(val) == len(self)
        self._value = val
        self._esd = [None] * len(self)

    @property
    def name_stem(self):
        return self._name_stem

    def __str__(self):

        msg = "ScanVaryingParameterSet " + self.name_stem + ":\n"
        try:
            msg += "    Type: " + self.param_type + "\n"
        except TypeError:
            msg += "    Type: " + str(self.param_type) + "\n"
        try:
            msg += "    Axis: (%5.3f, %5.3f, %5.3f)" % tuple(self.axis) + "\n"
        except TypeError:
            msg += "    Axis: " + str(self.axis) + "\n"
        vals = ", ".join(["%5.3f"] * len(self)) % tuple(self.value)
        msg += "    Values: " + vals + "\n"
        try:
            sigs = ", ".join(["%5.3f"] * len(self)) % tuple(self.esd)
        except TypeError:
            sigs = ", ".join([str(e) for e in self.esd])
        msg += "    Sigmas: (" + sigs + ") \n"

        return msg


# wrap the C++ GaussianSmoother, modifying return values to emulate the
# old Python version.
class GaussianSmoother(GS):
    """A Gaussian smoother for ScanVaryingModelParameterisations"""

    def value_weight(self, x, param):
        result = super().value_weight(x, flex.double(param.value))
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def multi_value_weight(self, x, param):
        result = super().multi_value_weight(flex.double(x), flex.double(param.value))
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def positions(self):
        return list(super().positions())


class ScanVaryingModelParameterisation(ModelParameterisation):
    """Extending ModelParameterisation to deal with ScanVaryingParameterSets.

    For simplicity at this stage it is decreed that a
    ScanVaryingModelParameterisation consists only of ScanVaryingParameterSets.
    There is no combination with normal Parameters. This could be changed later,
    but there may be no reason to do so, hence starting with this simpler
    design"""

    # The initial state is here equivalent to the initial state of the
    # time static version of the parameterisation, as it is assumed that we
    # start with a flat model wrt rotation angle.

    def __init__(
        self,
        model,
        initial_state,
        param_sets,
        smoother,
        experiment_ids,
        is_multi_state=False,
    ):

        ModelParameterisation.__init__(
            self, model, initial_state, param_sets, experiment_ids, is_multi_state
        )

        self._num_sets = len(self._param)
        self._num_samples = len(param_sets[0])
        self._total_len = self._num_samples * self._num_sets

        # ensure all internal parameter sets have the same number of parameters
        for param in self._param[1:]:
            assert len(param) == self._num_samples

        # Link up with an object that will perform the smoothing.
        self._smoother = smoother
        assert self._smoother.num_values() == self._num_samples

        # define an attribute for caching the variance-covariance matrix of
        # parameters
        self._var_cov = None

        return

    def num_samples(self):
        """the number of samples of each parameter"""
        return self._num_samples

    def num_free(self):
        """the number of free parameters"""

        if self._num_free is None:
            self._num_free = (
                sum(not x.get_fixed() for x in self._param) * self._num_samples
            )
        return self._num_free

    # def num_total(self): inherited unchanged from ModelParameterisation

    def num_sets(self):
        """the number of parameter sets"""
        return self._num_sets

    def compose(self, t):
        """compose the model state at image number t from its initial state and
        its parameter list. Also calculate the derivatives of the state wrt
        each parameter in the list.

        Unlike ModelParameterisation, does not automatically update the actual
        model class. This should be done once refinement is complete."""

        raise NotImplementedError()

    def get_param_vals(self, only_free=True):
        """export the values of the internal list of parameters as a
        sequence of floats.

        If only_free, the values of fixed parameters are filtered from the
        returned list. Otherwise all parameter values are returned"""

        if only_free:
            return [x for e in self._param if not e.get_fixed() for x in e.value]

        else:
            return [x for e in self._param for x in e.value]

    def get_param_names(self, only_free=True):
        """export the names of the internal list of parameters

        If only_free, the names of fixed parameters are filtered from the
        returned list. Otherwise all parameter names are returned"""

        if only_free:
            return [x for e in self._param if not e.get_fixed() for x in e.name]

        else:
            return [x for e in self._param for x in e.name]

    def set_param_vals(self, vals):
        """set the values of the internal list of parameters from a
        sequence of floats.

        First break the sequence into sub sequences of the same length
        as the _num_samples.

        Only free parameter sets can have values assigned, therefore the
        length of vals must equal the value of num_free"""

        assert len(vals) == self.num_free()
        i = 0
        for p in self._param:
            if not p.get_fixed():  # only set the free parameter sets
                new_vals = vals[i : i + self._num_samples]
                p.value = new_vals
                i += self._num_samples

        # compose with the new parameter values
        # self.compose()

        return

    def set_param_esds(self, esds):
        """set the estimated standard deviations of the internal list of parameters
        from a sequence of floats.

        First break the sequence into sub sequences of the same length
        as the _num_samples.

        Only free parameters can be set, therefore the length of esds must equal
        the value of num_free"""

        assert len(esds) == self.num_free()
        i = 0
        for p in self._param:
            if not p.get_fixed():  # only set the free parameter sets
                new_esds = esds[i : i + self._num_samples]
                p.esd = new_esds
                i += self._num_samples

        return

    # def get_fixed(self): inherited unchanged from ModelParameterisation

    # def set_fixed(self, fix): inherited unchanged from ModelParameterisation

    # def get_state(self): inherited unchanged from ModelParameterisation

    def get_ds_dp(self, only_free=True, use_none_as_null=False):
        """get a list of derivatives of the state wrt each parameter, as
        a list in the same order as the internal list of parameters. Requires
        compose to be called first at scan coordinate 't' so that each
        scan-dependent parameter is evaluated at coordinate t, corresponding to
        the original, unnormalised coordinates used to set up the smoother
        (t will most likely be along the dimension of image number).

        If only_free, the derivatives with respect to fixed parameters are
        omitted from the returned list. Otherwise a list for all parameters is
        returned, with null values for the fixed parameters.

        The internal list of derivatives self._dstate_dp may use None for null
        elements. By default these are converted to the null state, but
        optionally these may remain None to detect them easier and avoid
        doing calculations on null elements
        """

        if use_none_as_null:
            null = None
        else:
            null = self._null_state

        if only_free:
            return [
                null if ds_dp is None else ds_dp
                for row, p in zip(self._dstate_dp, self._param)
                if not p.get_fixed()
                for ds_dp in row
            ]

        else:
            return [
                null if p.get_fixed() or ds_dp is None else ds_dp
                for row, p in zip(self._dstate_dp, self._param)
                for ds_dp in row
            ]

    def get_smoothed_parameter_value(self, t, pset):
        """export the smoothed value of a parameter set at image number 't'
        using the smoother."""

        return self._smoother.value_weight(t, pset)[0]

    def calculate_state_uncertainties(self, var_cov=None):
        """Given a variance-covariance array for the parameters of this model,
        propagate those estimated errors into the uncertainties of the model state
        at every scan point"""

        if var_cov is not None:
            # first call, just cache the variance-covariance matrix
            self._var_cov = var_cov
            return None

        # later calls, make sure it has been cached! Otherwise ESDs cannot be
        # calculated, so return None
        if self._var_cov is None:
            return None

        # later calls, assumes compose has been called at image number t, so that
        # get_ds_dp will be specific for that image. Now call the base class method
        # and return the result
        return super().calculate_state_uncertainties(self._var_cov)

    def set_state_uncertainties(self, var_cov_list):
        """Send the calculated variance-covariance matrices for model state elements
        for all scan points back to the model for storage alongside the model state
        """

        pass



from __future__ import annotations

import bz2
import datetime
import gzip
import logging
import lzma
import math
import time

import iotbx.cif.model
from cctbx import crystal as cctbxcrystal
from cctbx import miller
from cctbx.sgtbx import bravais_types
from iotbx.merging_statistics import dataset_statistics
from libtbx import Auto
from scitbx.array_family import flex

import dials.util.version
from dials.util.filter_reflections import filter_reflection_table

logger = logging.getLogger(__name__)
RAD2DEG = 180.0 / math.pi


class MMCIFOutputFile:
    """
    Class to output experiments and reflections as MMCIF file
    """

    def __init__(self, params):
        """
        Init with the parameters
        """
        self._cif = iotbx.cif.model.cif()
        self.params = params
        self._v5_next_fmt = "%6i %2i %5i %5i %-2i %-2i %-2i"
        self._v5_0_fmt = "%2i %6i %-2i %-2i %-2i %6i %5.3f %5.3f"

    def write(self, experiments, reflections):
        """
        Write the experiments and reflections to file
        """

        # if mmcif filename is auto, then choose scaled.cif or integrated.cif
        if self.params.mmcif.hklout in (None, Auto, "auto"):
            if ("intensity.scale.value" in reflections) and (
                "intensity.scale.variance" in reflections
            ):
                filename = "scaled.cif"
                logger.info(
                    "Data appears to be scaled, setting mmcif.hklout = 'scaled.cif'"
                )
            else:
                filename = "integrated.cif"
                logger.info(
                    "Data appears to be unscaled, setting mmcif.hklout = 'integrated.cif'"
                )
        else:
            filename = self.params.mmcif.hklout

        # Add the block
        self._cif["dials"] = self.make_cif_block(experiments, reflections)

        if self.params.mmcif.pdb_version == "v5":
            loop_format_strings = {"_diffrn_refln": self._v5_0_fmt}
        else:
            loop_format_strings = {"_pdbx_diffrn_unmerged_refln": self._v5_next_fmt}

        # Print to file
        if self.params.mmcif.compress and not filename.endswith(
            "." + self.params.mmcif.compress
        ):
            filename += "." + self.params.mmcif.compress
        if self.params.mmcif.compress == "gz":
            open_fn = gzip.open
        elif self.params.mmcif.compress == "bz2":
            open_fn = bz2.open
        elif self.params.mmcif.compress == "xz":
            open_fn = lzma.open
        else:
            open_fn = open
        with open_fn(filename, "wt") as fh:
            self._cif.show(out=fh, loop_format_strings=loop_format_strings)

        # Log
        logger.info("Wrote reflections to %s", filename)

    def make_cif_block(self, experiments, reflections):
        """Write the data to a cif block"""
        # Select reflections
        selection = reflections.get_flags(reflections.flags.integrated, all=True)
        reflections = reflections.select(selection)

        # Filter out bad variances and other issues, but don't filter on ice rings
        # or alter partialities.

        # Assumes you want to apply the lp and dqe corrections to sum and prf
        # Do we want to combine partials?
        reflections = filter_reflection_table(
            reflections,
            self.params.intensity,
            combine_partials=False,
            partiality_threshold=0.0,
            d_min=self.params.mtz.d_min,
        )

        # Get the cif block
        cif_block = iotbx.cif.model.block()

        # Audit trail
        dials_version = dials.util.version.dials_version()
        cif_block["_audit.revision_id"] = 1
        cif_block["_audit.creation_method"] = dials_version
        cif_block["_audit.creation_date"] = datetime.date.today().isoformat()
        cif_block["_entry.id"] = "DIALS"
        # add software loop
        mmcif_software_header = (
            "_software.pdbx_ordinal",
            "_software.citation_id",
            "_software.name",  # as defined at [1]
            "_software.version",
            "_software.type",
            "_software.classification",
            "_software.description",
        )

        mmcif_citations_header = (
            "_citation.id",
            "_citation.journal_abbrev",
            "_citation.journal_volume",
            "_citation.journal_issue",
            "_citation.page_first",
            "_citation.page_last",
            "_citation.year",
            "_citation.title",
        )

        software_loop = iotbx.cif.model.loop(header=mmcif_software_header)
        citations_loop = iotbx.cif.model.loop(header=mmcif_citations_header)

        software_loop.add_row(
            (
                1,
                1,
                "DIALS",
                dials_version,
                "package",
                "data processing",
                "Data processing and integration within the DIALS software package",
            )
        )
        citations_loop.add_row(
            (
                1,
                "Acta Cryst. D",
                74,
                2,
                85,
                97,
                2018,
                "DIALS: implementation and evaluation of a new integration package",
            )
        )
        if "scale" in self.params.intensity:
            software_loop.add_row(
                (
                    2,
                    2,
                    "DIALS",
                    dials_version,
                    "program",
                    "data scaling",
                    "Data scaling and merging within the DIALS software package",
                )
            )
            citations_loop.add_row(
                (
                    2,
                    "Acta Cryst. D",
                    76,
                    4,
                    385,
                    399,
                    2020,
                    "Scaling diffraction data in the DIALS software package: algorithms and new approaches for multi-crystal scaling",
                )
            )
        cif_block.add_loop(software_loop)
        cif_block.add_loop(citations_loop)

        # Hard coding X-ray
        if self.params.mmcif.pdb_version == "v5_next":
            cif_block["_pdbx_diffrn_data_section.id"] = "dials"
            cif_block["_pdbx_diffrn_data_section.type_scattering"] = "x-ray"
            cif_block["_pdbx_diffrn_data_section.type_merged"] = "false"
            cif_block["_pdbx_diffrn_data_section.type_scaled"] = str(
                "scale" in self.params.intensity
            ).lower()

        # FIXME finish metadata addition - detector and source details needed
        # http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Categories/index.html

        # Add source information;
        # _diffrn_source.pdbx_wavelength_list = (list of wavelengths)
        # _diffrn_source.source = (general class of source e.g. synchrotron)
        # _diffrn_source.type = (specific beamline or instrument e.g DIAMOND BEAMLINE I04)

        wls = []
        epochs = []
        for exp in experiments:
            wls.append(round(exp.beam.get_wavelength(), 5))
            epochs.append(exp.scan.get_epochs()[0])
        unique_wls = set(wls)
        cif_block["_exptl_crystal.id"] = 1  # links to crystal_id
        cif_block["_diffrn.id"] = 1  # links to diffrn_id
        cif_block["_diffrn.crystal_id"] = 1
        cif_block["_diffrn_source.diffrn_id"] = 1
        cif_block["_diffrn_source.pdbx_wavelength_list"] = ", ".join(
            str(w) for w in unique_wls
        )

        # Add detector information;
        # _diffrn_detector.detector  = (general class e.g. PIXEL, PLATE etc)
        # _diffrn_detector.pdbx_collection_date = (Date of collection yyyy-mm-dd)
        # _diffrn_detector.type = (full name of detector e.g. DECTRIS PILATUS3 2M)
        # One date is required, so if multiple just use the first date.
        min_epoch = min(epochs)
        date_str = time.strftime("%Y-%m-%d", time.gmtime(min_epoch))
        cif_block["_diffrn_detector.diffrn_id"] = 1
        cif_block["_diffrn_detector.pdbx_collection_date"] = date_str

        # Write reflection data
        # Required columns
        v5_next_header = (
            "_pdbx_diffrn_unmerged_refln.reflection_id",
            "_pdbx_diffrn_unmerged_refln.scan_id",
            "_pdbx_diffrn_unmerged_refln.image_id_begin",
            "_pdbx_diffrn_unmerged_refln.image_id_end",
            "_pdbx_diffrn_unmerged_refln.index_h",
            "_pdbx_diffrn_unmerged_refln.index_k",
            "_pdbx_diffrn_unmerged_refln.index_l",
        )

        v5_0_header = (
            "_diffrn_refln.diffrn_id",  # pointer to _diffrn.id in the DIFFRN category.
            "_diffrn_refln.id",  # uniquely identifies reflection
            "_diffrn_refln.index_h",
            "_diffrn_refln.index_k",
            "_diffrn_refln.index_l",
            "_diffrn_refln.pdbx_image_id",
            "_diffrn_refln.pdbx_detector_x",
            "_diffrn_refln.pdbx_detector_y",
        )

        v5_next_extra_items = {
            "scales": ("_pdbx_diffrn_unmerged_refln.scale_value", "%5.3f"),
            "intensity.scale.value": (
                "_pdbx_diffrn_unmerged_refln.intensity_meas",
                "%8.3f",
            ),
            "intensity.scale.sigma": (
                "_pdbx_diffrn_unmerged_refln.intensity_sigma",
                "%8.3f",
            ),
            "intensity.sum.value": (
                "_pdbx_diffrn_unmerged_refln.intensity_sum",
                "%8.3f",
            ),
            "intensity.sum.sigma": (
                "_pdbx_diffrn_unmerged_refln.intensity_sum_sigma",
                "%8.3f",
            ),
            "intensity.prf.value": (
                "_pdbx_diffrn_unmerged_refln.intensity_prf",
                "%8.3f",
            ),
            "intensity.prf.sigma": (
                "_pdbx_diffrn_unmerged_refln.intensity_prf_sigma",
                "%8.3f",
            ),
            "angle": ("_pdbx_diffrn_unmerged_refln.scan_angle_reflection", "%7.4f"),
            "partiality": ("_pdbx_diffrn_unmerged_refln.partiality", "%7.4f"),
        }

        v5_0_extra_items = {
            "scales": ("_diffrn_refln.pdbx_scale_value", "%5.3f"),
            "intensity.scale.value": ("_diffrn_refln.intensity_net", "%8.3f"),
            "intensity.scale.sigma": ("_diffrn_refln.intensity_sigma", "%8.3f"),
            "angle": ("_diffrn_refln.pdbx_scan_angle", "%7.4f"),
        }

        variables_present = []
        if "scale" in self.params.intensity:
            reflections["scales"] = 1.0 / reflections["inverse_scale_factor"]
            reflections["intensity.scale.sigma"] = flex.sqrt(
                reflections["intensity.scale.variance"]
            )
            variables_present.extend(
                ["scales", "intensity.scale.value", "intensity.scale.sigma"]
            )
        if self.params.mmcif.pdb_version == "v5_next":
            if "sum" in self.params.intensity:
                reflections["intensity.sum.sigma"] = flex.sqrt(
                    reflections["intensity.sum.variance"]
                )
                variables_present.extend(["intensity.sum.value", "intensity.sum.sigma"])
            if "profile" in self.params.intensity:
                reflections["intensity.prf.sigma"] = flex.sqrt(
                    reflections["intensity.prf.variance"]
                )
                variables_present.extend(["intensity.prf.value", "intensity.prf.sigma"])

        # Should always exist
        reflections["angle"] = reflections["xyzcal.mm"].parts()[2] * RAD2DEG
        variables_present.extend(["angle"])

        if self.params.mmcif.pdb_version == "v5_next":
            if "partiality" in reflections:
                variables_present.extend(["partiality"])

        for name in variables_present:
            if name in reflections:
                v5_next_header += (v5_next_extra_items[name][0],)
                self._v5_next_fmt += " " + v5_next_extra_items[name][1]

        for name in variables_present:
            if name in reflections and name in v5_0_extra_items:
                v5_0_header += (v5_0_extra_items[name][0],)
                self._v5_0_fmt += " " + v5_0_extra_items[name][1]

        if "scale" in self.params.intensity:
            # Write dataset_statistics - first make a miller array
            crystal_symmetry = cctbxcrystal.symmetry(
                space_group=experiments[0].crystal.get_space_group(),
                unit_cell=experiments[0].crystal.get_unit_cell(),
            )
            miller_set = miller.set(
                crystal_symmetry=crystal_symmetry,
                indices=reflections["miller_index"],
                anomalous_flag=False,
            )
            i_obs = miller.array(miller_set, data=reflections["intensity.scale.value"])
            i_obs.set_observation_type_xray_intensity()
            i_obs.set_sigmas(reflections["intensity.scale.sigma"])
            i_obs.set_info(
                miller.array_info(source="DIALS", source_type="reflection_tables")
            )

            result = dataset_statistics(
                i_obs=i_obs,
                crystal_symmetry=crystal_symmetry,
                use_internal_variance=False,
                eliminate_sys_absent=False,
                assert_is_not_unique_set_under_symmetry=False,
            )
            merged_block = iotbx.cif.model.block()
            merged_block["_reflns.pdbx_ordinal"] = 1
            merged_block["_reflns.pdbx_diffrn_id"] = 1
            merged_block["_reflns.entry_id"] = "DIALS"
            merged_data = result.as_cif_block()
            merged_block.update(merged_data)
            cif_block.update(merged_block)

        # Write the crystal information
        # if v5, that's all so return
        if self.params.mmcif.pdb_version == "v5":
            h, k, l = [
                hkl.iround()
                for hkl in reflections["miller_index"].as_vec3_double().parts()
            ]
            # Note, use observed position, so that we are within the
            # allowed bounds (lower bound 0) for image_id
            det_x, det_y, det_z = reflections["xyzobs.px.value"].parts()
            image_id = flex.ceil(det_z).iround()
            loop_values = [
                flex.int(len(reflections), 1),  # diffn id
                flex.size_t_range(1, len(reflections) + 1),  # refln id
                h,
                k,
                l,
                image_id,
                det_x,
                det_y,
            ] + [reflections[name] for name in variables_present]
            cif_loop = iotbx.cif.model.loop(data=dict(zip(v5_0_header, loop_values)))
            cif_block.add_loop(cif_loop)

            return cif_block
        # continue if v5_next
        cif_loop = iotbx.cif.model.loop(
            header=(
                "_pdbx_diffrn_unmerged_cell.ordinal",
                "_pdbx_diffrn_unmerged_cell.crystal_id",
                "_pdbx_diffrn_unmerged_cell.wavelength",
                "_pdbx_diffrn_unmerged_cell.cell_length_a",
                "_pdbx_diffrn_unmerged_cell.cell_length_b",
                "_pdbx_diffrn_unmerged_cell.cell_length_c",
                "_pdbx_diffrn_unmerged_cell.cell_angle_alpha",
                "_pdbx_diffrn_unmerged_cell.cell_angle_beta",
                "_pdbx_diffrn_unmerged_cell.cell_angle_gamma",
                "_pdbx_diffrn_unmerged_cell.Bravais_lattice",
            )
        )
        crystals = experiments.crystals()
        crystal_to_id = {crystal: i + 1 for i, crystal in enumerate(crystals)}
        for i, exp in enumerate(experiments):
            crystal = exp.crystal
            crystal_id = crystal_to_id[crystal]
            wavelength = exp.beam.get_wavelength()
            a, b, c, alpha, beta, gamma = crystal.get_unit_cell().parameters()
            latt_type = str(
                bravais_types.bravais_lattice(group=crystal.get_space_group())
            )
            cif_loop.add_row(
                (i + 1, crystal_id, wavelength, a, b, c, alpha, beta, gamma, latt_type)
            )
            cif_block.add_loop(cif_loop)

        # Write the scan information
        cif_loop = iotbx.cif.model.loop(
            header=(
                "_pdbx_diffrn_scan.scan_id",
                "_pdbx_diffrn_scan.crystal_id",
                "_pdbx_diffrn_scan.image_id_begin",
                "_pdbx_diffrn_scan.image_id_end",
                "_pdbx_diffrn_scan.scan_angle_begin",
                "_pdbx_diffrn_scan.scan_angle_end",
            )
        )

        expid_to_scan_id = {exp.identifier: i + 1 for i, exp in enumerate(experiments)}

        for i, exp in enumerate(experiments):
            scan = exp.scan
            crystal_id = crystal_to_id[exp.crystal]
            image_range = scan.get_image_range()
            osc_range = scan.get_oscillation_range(deg=True)
            cif_loop.add_row(
                (
                    i + 1,
                    crystal_id,
                    image_range[0],
                    image_range[1],
                    osc_range[0],
                    osc_range[1],
                )
            )
            cif_block.add_loop(cif_loop)

        _, _, _, _, z0, z1 = reflections["bbox"].parts()
        h, k, l = [
            hkl.iround() for hkl in reflections["miller_index"].as_vec3_double().parts()
        ]
        # make scan id consistent with header as defined above
        scan_id = flex.int(reflections.size(), 0)
        for id_ in reflections.experiment_identifiers().keys():
            expid = reflections.experiment_identifiers()[id_]
            sel = reflections["id"] == id_
            scan_id.set_selected(sel, expid_to_scan_id[expid])

        loop_values = [
            flex.size_t_range(1, len(reflections) + 1),
            scan_id,
            z0,
            z1,
            h,
            k,
            l,
        ] + [reflections[name] for name in variables_present]
        cif_loop = iotbx.cif.model.loop(data=dict(zip(v5_next_header, loop_values)))
        cif_block.add_loop(cif_loop)

        return cif_block



"""
tests for functions in dials.util.exclude_images.py
"""

from __future__ import annotations

import copy
from unittest.mock import Mock

import pytest

from dxtbx.model import Experiment, ExperimentList, Scan

from dials.array_family import flex
from dials.util.exclude_images import (
    _parse_exclude_images_commands,
    exclude_image_ranges_for_scaling,
    exclude_image_ranges_from_scans,
    get_selection_for_valid_image_ranges,
    get_valid_image_ranges,
    set_initial_valid_image_ranges,
)


def make_scan_experiment(image_range=(1, 100), expid="0"):
    """Make an experiment with a scan"""
    return Experiment(scan=Scan(image_range, (0.0, 1.0)), identifier=expid)


def make_scanless_experiment(expid="1"):
    """Make an experiment without a scan i.e. single-image dataset"""
    return Experiment(identifier=expid)


def test_parse_exclude_images_commands():
    """Test for namesake function"""
    formats = (
        [["1:101:200"], ["0:201:300"]],  # if given as separate exclude_images=
        [["1:101:200,0:201:300"]],  # if given as exclude_images="1:101:200,0:201:300"
        [
            ["1:101:200", "0:201:300"]
        ],  # if given as exclude_images="1:101:200 0:201:300"
    )
    for command in formats:
        r1 = flex.reflection_table()
        r1.experiment_identifiers()[1] = "1"
        r0 = flex.reflection_table()
        r0.experiment_identifiers()[0] = "0"
        tables = [r0, r1]
        ranges = _parse_exclude_images_commands(command, [], tables)
        assert ranges == [("1", (101, 200)), ("0", (201, 300))]

    experiments = ["1", "2"]
    short_command = [["101:200"]]
    with pytest.raises(ValueError):
        _ = _parse_exclude_images_commands(short_command, experiments, tables)
    mock_exp = Mock()
    mock_exp.identifier = "1"
    ranges = _parse_exclude_images_commands(short_command, [mock_exp], tables)
    assert ranges == [("1", (101, 200))]
    with pytest.raises(ValueError):
        _ = _parse_exclude_images_commands([["1:101-200"]], [mock_exp], tables)
    with pytest.raises(ValueError):
        _ = _parse_exclude_images_commands([["1:101:a"]], [], tables)


def test_set_get_initial_valid_image_ranges():
    """Test for get/set valid_image_ranges functions"""
    explist = ExperimentList([make_scan_experiment(), make_scanless_experiment()])
    explist = set_initial_valid_image_ranges(explist)
    assert list(explist[0].scan.get_valid_image_ranges("0")) == [(1, 100)]
    ranges = get_valid_image_ranges(explist)
    assert len(ranges) == 2
    assert list(ranges[0]) == [(1, 100)]
    assert ranges[1] is None


def test_exclude_image_ranges_from_scans():
    """Test for namesake function"""
    explist = ExperimentList(
        [make_scan_experiment(expid="0"), make_scan_experiment(expid="1")]
    )
    exclude_images = [["0:81:100"], ["1:61:80"]]
    r1 = flex.reflection_table()
    r1.experiment_identifiers()[1] = "1"
    r0 = flex.reflection_table()
    r0.experiment_identifiers()[0] = "0"
    tables = [r0, r1]
    explist = exclude_image_ranges_from_scans(tables, explist, exclude_images)
    assert list(explist[0].scan.get_valid_image_ranges("0")) == [(1, 80)]
    assert list(explist[1].scan.get_valid_image_ranges("1")) == [(1, 60), (81, 100)]
    # Try excluding a range that already has been excluded
    explist = exclude_image_ranges_from_scans(tables, explist, [["1:70:80"]])
    assert list(explist[0].scan.get_valid_image_ranges("0")) == [(1, 80)]
    assert list(explist[1].scan.get_valid_image_ranges("1")) == [(1, 60), (81, 100)]
    scanlessexplist = ExperimentList([make_scanless_experiment()])
    with pytest.raises(ValueError):
        _ = exclude_image_ranges_from_scans(tables, scanlessexplist, [["0:1:100"]])
    # Now try excluding everything, should set an empty array
    explist = exclude_image_ranges_from_scans(tables, explist, [["1:1:100"]])
    assert list(explist[0].scan.get_valid_image_ranges("0")) == [(1, 80)]
    assert list(explist[1].scan.get_valid_image_ranges("1")) == []

    ## test what happens if a single image is left within the scan
    explist = ExperimentList(
        [make_scan_experiment(expid="0"), make_scan_experiment(expid="1")]
    )
    exclude_images = [["0:81:100"], ["1:76:79"], ["1:81:99"]]
    r1 = flex.reflection_table()
    r1.experiment_identifiers()[1] = "1"
    r0 = flex.reflection_table()
    r0.experiment_identifiers()[0] = "0"
    tables = [r0, r1]
    explist = exclude_image_ranges_from_scans(tables, explist, exclude_images)
    assert list(explist[0].scan.get_valid_image_ranges("0")) == [(1, 80)]
    assert list(explist[1].scan.get_valid_image_ranges("1")) == [
        (1, 75),
        (80, 80),
        (100, 100),
    ]


def test_get_selection_for_valid_image_ranges():
    """Test for namesake function"""
    exp = make_scan_experiment()
    exp.scan.set_valid_image_ranges("0", [(2, 10)])
    refl = flex.reflection_table()
    refl["xyzobs.px.value"] = flex.vec3_double(
        [(0, 0, 0.5), (0, 0, 1.5), (0, 0, 5.5), (0, 0, 9.5), (0, 0, 10.5)]
    )
    sel = get_selection_for_valid_image_ranges(refl, exp)
    assert list(sel) == [False, True, True, True, False]
    exp = make_scanless_experiment()
    assert list(get_selection_for_valid_image_ranges(refl, exp)) == [True] * 5


def test_exclude_image_ranges_for_scaling():
    """Test for namesake function."""
    refl1 = flex.reflection_table()
    refl1["xyzobs.px.value"] = flex.vec3_double(
        [(0, 0, 0.5), (0, 0, 1.5), (0, 0, 5.5), (0, 0, 9.5), (0, 0, 10.5)]
    )
    refl1.set_flags(flex.bool(5, False), refl1.flags.user_excluded_in_scaling)
    refl2 = copy.deepcopy(refl1)
    refl1.experiment_identifiers()[0] = "0"
    refl2.experiment_identifiers()[1] = "1"
    explist = ExperimentList(
        [
            make_scan_experiment(image_range=(2, 20), expid="0"),
            make_scan_experiment(image_range=(2, 20), expid="1"),
        ]
    )
    refls, explist = exclude_image_ranges_for_scaling(
        [refl1, refl2], explist, [["1:11:20"]]
    )
    assert list(explist[0].scan.get_valid_image_ranges("0")) == [(2, 20)]
    assert list(explist[1].scan.get_valid_image_ranges("1")) == [(2, 10)]
    assert list(refls[0].get_flags(refls[0].flags.user_excluded_in_scaling)) == [
        True,
        False,
        False,
        False,
        False,
    ]
    assert list(refls[1].get_flags(refls[0].flags.user_excluded_in_scaling)) == [
        True,
        False,
        False,
        False,
        True,
    ]



from __future__ import annotations

import concurrent.futures
import logging
import math

import numpy as np

import libtbx
from cctbx import sgtbx, uctbx
from iotbx import ccp4_map, phil
from scitbx.array_family import flex

import dials.algorithms.rs_mapper as recviewer
import dials.util
import dials.util.log
from dials.util import Sorry
from dials.util.mp import available_cores
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """
This program reconstructs reciprocal space from diffraction images. The orientation matrix is not necessary; only diffraction geometry is required.

This program is intended to help detection and visualization of pathologies such as multiple-lattice, twinning, modulation, diffuse scattering and high background. It is also useful for education.

Examples::

  dials.rs_mapper image1.cbf

  dials.rs_mapper image_00*.cbf

  dials.rs_mapper imported.expt
"""

# Define a logger
logger = logging.getLogger("dials.rs_mapper")

phil_scope = phil.parse(
    """
rs_mapper
  .short_caption = Reciprocal space mapper
{
  max_resolution = 6
    .type = float
    .optional = True
    .short_caption = Resolution limit
  grid_size = 192
    .type = int
    .optional = True
  reverse_phi = False
    .type = bool
    .optional = True
  ignore_mask = False
    .type = bool
    .optional = True
    .short_caption = Ignore masks from dxtbx class
  nproc = Auto
    .help = "Number of processes over which to split the calculation. If set to"
            "Auto, DIALS will choose automatically."
    .type = int(value_min=1)
    .expert_level = 1
}
output
{
  map_file = rs_mapper_output.mrc
    .type = path
    .optional = False
    .multiple= False
    .short_caption = Output map file

  log = "dials.rs_mapper.log"
    .type = str
}
""",
    process_includes=True,
)


def process_block(
    block, imageset, grid_size, reverse_phi, S, ignore_mask, xy, rec_range
):
    grid = flex.double(flex.grid(grid_size, grid_size, grid_size), 0)
    counts = flex.int(flex.grid(grid_size, grid_size, grid_size), 0)

    axis = imageset.get_goniometer().get_rotation_axis()
    for i in block:
        osc_range = imageset.get_scan(i).get_oscillation_range()

        angle = (osc_range[0] + osc_range[1]) / 2 / 180 * math.pi
        if not reverse_phi:
            # the pixel is in S AFTER rotation. Thus we have to rotate BACK.
            angle *= -1
        rotated_S = S.rotate_around_origin(axis, angle)

        data = imageset.get_raw_data(i)[0]
        if not ignore_mask:
            mask = imageset.get_mask(i)[0]
            data.set_selected(~mask, 0)

        recviewer.fill_voxels(data, grid, counts, rotated_S, xy, rec_range)

    return grid, counts


class Script:
    def __init__(self):
        """Initialise the script."""

        # The script usage
        usage = (
            "usage: dials.rs_mapper map_file=output.ccp4 [max_resolution=6] [grid_size=192] "
            "[reverse_phi=False] [param.phil] "
            "{image1.file [image2.file ...]} | imported.expt"
        )

        # Initialise the base class
        self.parser = ArgumentParser(
            usage=usage, phil=phil_scope, epilog=help_message, read_experiments=True
        )

    def run(self, args=None):
        # Parse the command line
        params, options = self.parser.parse_args(args, show_diff_phil=True)

        # Configure the logging.
        dials.util.log.config(options.verbose, logfile=params.output.log)

        if not params.output.map_file:
            raise RuntimeError("Please specify output map file (map_file=)")
        else:
            self.map_file = params.output.map_file

        # Ensure we have either a data block or an experiment list
        self.experiments = flatten_experiments(params.input.experiments)
        if len(self.experiments) != 1:
            self.parser.print_help()
            return

        self.reverse_phi = params.rs_mapper.reverse_phi
        self.grid_size = params.rs_mapper.grid_size
        self.max_resolution = params.rs_mapper.max_resolution
        self.ignore_mask = params.rs_mapper.ignore_mask

        self.grid = flex.double(
            flex.grid(self.grid_size, self.grid_size, self.grid_size), 0
        )
        self.counts = flex.int(
            flex.grid(self.grid_size, self.grid_size, self.grid_size), 0
        )

        self.nproc = params.rs_mapper.nproc
        if self.nproc is libtbx.Auto:
            self.nproc = available_cores()
            logger.info("Setting nproc={}".format(self.nproc))

        for experiment in self.experiments:
            grid, counts = self.process_imageset(experiment.imageset)

            self.grid += grid
            self.counts += counts

        recviewer.normalize_voxels(self.grid, self.counts)

        # Let's use 1/(100A) as the unit so that the absolute numbers in the
        # "cell dimensions" field of the ccp4 map are typical for normal
        # MX maps. The values in 1/A would give the "cell dimensions" around
        # or below 1 and some MX programs would not handle it well.
        box_size = 100 * 2.0 / self.max_resolution
        uc = uctbx.unit_cell((box_size, box_size, box_size, 90, 90, 90))
        logger.info(f"Saving map to {self.map_file}")
        ccp4_map.write_ccp4_map(
            self.map_file,
            uc,
            sgtbx.space_group("P1"),
            (0, 0, 0),
            self.grid.all(),
            self.grid,
            flex.std_string(["cctbx.miller.fft_map"]),
        )

    def process_imageset(self, imageset):
        rec_range = 1 / self.max_resolution

        if len(imageset.get_detector()) != 1:
            raise Sorry("This program does not support multi-panel detectors.")

        panel = imageset.get_detector()[0]
        beam = imageset.get_beam()
        s0 = beam.get_s0()
        pixel_size = panel.get_pixel_size()
        xlim, ylim = imageset.get_raw_data(0)[0].all()
        if pixel_size[0] != pixel_size[1]:
            raise Sorry("This program does not support non-square pixels.")

        # cache transformation
        xy = recviewer.get_target_pixels(panel, s0, xlim, ylim, self.max_resolution)
        s1 = panel.get_lab_coord(xy * pixel_size[0])
        s1 = s1 / s1.norms() * (1 / beam.get_wavelength())
        S = s1 - s0

        # Split imageset into up to nproc blocks of at least 10 images
        nblocks = min(self.nproc, int(math.ceil(len(imageset) / 10)))
        blocks = np.array_split(range(len(imageset)), nblocks)
        blocks = [block.tolist() for block in blocks]

        logger.info(f"Calculation split over {len(blocks)} blocks")
        header = ["Block", "Oscillation range (°)"]
        scan = imageset.get_scan()
        rows = [
            [
                f"{i+1}",
                f"{scan.get_angle_from_array_index(block[0]):.2f} - {scan.get_angle_from_array_index(block[-1] + 1):.2f}",
            ]
            for i, block in enumerate(blocks)
        ]
        logger.info(dials.util.tabulate(rows, header, numalign="right") + "\n")

        if len(blocks) == 1:
            results = [
                process_block(
                    blocks[0],
                    imageset,
                    self.grid_size,
                    self.reverse_phi,
                    S,
                    self.ignore_mask,
                    xy,
                    rec_range,
                ),
            ]
        else:
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=len(blocks)
            ) as pool:
                results = [
                    pool.submit(
                        process_block,
                        block,
                        imageset,
                        self.grid_size,
                        self.reverse_phi,
                        S,
                        self.ignore_mask,
                        xy,
                        rec_range,
                    )
                    for block in blocks
                ]
            results = [e.result() for e in results]

        grid, counts = results[0]
        for (g, c) in results[1:]:
            grid += g
            counts += c

        return grid, counts


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

import iotbx.phil
from scitbx import matrix

import dials.util

help_message = """
dials.goniometer_calibration is a tool to aid calibration of multi-axis
goniometers.

The tool takes as input experiments.expt files for datasets recorded at the
goniometer datum setting and for each goniometer axis incremented in turn. It
outputs the axes and angles relating each consecutive pair of crystal setting
matrices in imgCIF and MOSFLM coordinate systems, and the CIF loop describing
the goniometer axes. Optionally it can also output an XOalign configuration file.

Either space_group must be specified or the parameter
use_space_group_from_experiments=True must be set

Examples::

dials.goniometer_calibration space_group=P422 \\
  experiments_o0_k0_p0.expt experiments_o0_k0_p48.expt \\
  experiments_o0_k48_p48.expt experiments_o48_k48_p48.expt
"""


phil_scope = iotbx.phil.parse(
    """
space_group = None
  .type = space_group
use_space_group_from_experiments = False
  .type = bool
output {
  xoalign = None
    .type = path
}
"""
)


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util.options import ArgumentParser, flatten_experiments

    usage = "dials.goniometer_calibration [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    if not params.use_space_group_from_experiments and params.space_group is None:
        parser.print_help()
        return

    experiments = flatten_experiments(params.input.experiments)
    if len(experiments) <= 1:
        parser.print_help()
        return

    from dials.algorithms.indexing.compare_orientation_matrices import (
        difference_rotation_matrix_axis_angle,
    )

    for experiment in experiments:
        crystal = experiment.crystal
        gonio = experiment.goniometer
        assert len(experiments) == (len(gonio.get_axes()) + 1)
        scan = experiment.scan
        fixed_rotation = matrix.sqr(gonio.get_fixed_rotation())
        setting_rotation = matrix.sqr(gonio.get_setting_rotation())
        rotation_axis = matrix.col(gonio.get_rotation_axis_datum())
        rotation_matrix = rotation_axis.axis_and_angle_as_r3_rotation_matrix(
            scan.get_oscillation()[0], deg=True
        )
        U = matrix.sqr(crystal.get_U())
        U = setting_rotation * rotation_matrix * fixed_rotation * U
        crystal.set_U(U)
        if params.space_group is not None:
            crystal.set_space_group(params.space_group.group())

    rows = []

    from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame

    R_to_mosflm = align_reference_frame(
        experiments[0].beam.get_s0(),
        (1.0, 0.0, 0.0),
        experiments[0].goniometer.get_rotation_axis(),
        (0.0, 0.0, 1.0),
    )

    axes = []
    angles = []

    for i in range(len(experiments) - 1):
        target_angle = experiments[i + 1].goniometer.get_angles()[i]
        if i == experiments[i].goniometer.get_scan_axis():
            # rotation axis is canonical in our coordinate system
            axis = experiments[i].goniometer.get_axes()[i]
            angle = target_angle
        else:
            R_ij, axis, angle, cb_op = difference_rotation_matrix_axis_angle(
                experiments[i].crystal,
                experiments[i + 1].crystal,
                target_angle=target_angle,
            )
        gonio = experiments[i + 1].goniometer
        axis_names = gonio.get_names()
        axes.append(axis)
        angles.append(angle)
        depends_on = "."
        if i + 1 < len(axis_names):
            depends_on = axis_names[i + 1]
        rows.insert(
            0,
            (
                axis_names[i],
                "rotation",
                "goniometer",
                depends_on,
                f"{axis[0]:.4f}",
                f"{axis[1]:.4f}",
                f"{axis[2]:.4f}",
                ".",
                ".",
                ".",
            ),
        )

    axis_names = experiments[0].goniometer.get_names()
    print("Goniometer axes and angles (ImgCIF coordinate system):")
    for axis, angle, name in zip(axes, angles, axis_names):
        print(
            f"{name}: ",
            f"rotation of {angle:.3f} degrees",
            "about axis (%.5f,%.5f,%.5f)" % axis,
        )

    print()
    print("Goniometer axes and angles (MOSFLM coordinate system):")
    for axis, angle, name in zip(axes, angles, axis_names):
        print(
            f"{name}: ",
            f"rotation of {angle:.3f} degrees",
            "about axis (%.5f,%.5f,%.5f)" % (R_to_mosflm * matrix.col(axis)).elems,
        )

    print()
    print("ImgCIF _axis loop template:")
    from iotbx import cif

    loop = cif.model.loop(
        header=[
            "_axis.id",
            "_axis.type",
            "_axis.equipment",
            "_axis.depends_on",
            "_axis.vector[1]",
            "_axis.vector[2]",
            "_axis.vector[3]",
            "_axis.offset[1]",
            "_axis.offset[2]",
            "_axis.offset[3]",
        ]
    )
    for row in rows:
        loop.add_row(row)

    print(loop)

    if params.output.xoalign is not None:
        axes_mosflm = [(R_to_mosflm * matrix.col(axis)).elems for axis in axes]
        write_xoalign_config(
            params.output.xoalign, reversed(axes_mosflm), reversed(axis_names)
        )


def write_xoalign_config(file_name, axes, names):
    with open(file_name, "w") as f:
        print("GONIOMETER_AXES_NAMES = " + str(tuple(names)), file=f)
        print(
            "GONIOMETER_AXES = "
            + "["
            + ", ".join(
                "(" + ", ".join("%.5f" % x for x in axis) + ")" for axis in axes
            )
            + "]",
            file=f,
        )
        print("GONIOMETER_DATUM = (0,0,0) # in degrees", file=f)


if __name__ == "__main__":
    run()
